package no.ntnu.trec.kba;

import static org.apache.lucene.document.Field.Index.ANALYZED;
import static org.apache.lucene.document.Field.Index.NOT_ANALYZED_NO_NORMS;
import static org.apache.lucene.document.Field.TermVector.NO;
import static org.apache.lucene.document.Field.TermVector.WITH_POSITIONS_OFFSETS;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.lang.reflect.Array;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import no.ntnu.utils.NlpHelper;
import org.apache.thrift.TException;
import org.apache.thrift.protocol.TBinaryProtocol;
import org.apache.thrift.protocol.TProtocol;
import org.apache.thrift.transport.AutoExpandingBufferWriteTransport;
import org.apache.thrift.transport.TIOStreamTransport;
import streamcorpus.*;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;


/**
 * Document representation of a StreamItem
 * 
 * @author kbalog
 */
public class StreamDocument implements Writable {

	protected HashMap<String, String> fields;
	protected List<Sentence> sentences;

	public final static String FIELD_DOC_ID = "doc_id";
	public final static String FIELD_STREAM_ID = "stream_id";
	public final static String FIELD_DIR_NAME = "dir_name";
	public final static String FIELD_SOURCE = "source"; // this was called genre
														// before
	public final static String FIELD_SCHOST = "schost";
	public final static String FIELD_EPOCH_TICKS = "epoch_ticks";
	public final static String FIELD_ZULU_TIMESTAMP = "zulu_timestamp";
	public final static String FIELD_TITLE = "title_cleansed";
	public final static String FIELD_BODY = "body_cleansed";
	public final static String FIELD_ANCHOR = "anchor_cleansed";
	public final static String SERIF_TAGS = "serif_tags";

	/**
	 * Field names
	 */
	protected String[] fieldNames = { FIELD_DOC_ID, FIELD_STREAM_ID,
			FIELD_DIR_NAME, FIELD_SOURCE, FIELD_SCHOST, FIELD_EPOCH_TICKS,
			FIELD_ZULU_TIMESTAMP, SERIF_TAGS, FIELD_TITLE, FIELD_BODY, FIELD_ANCHOR };

	/**
	 * Lucene field indexing -- parallel array to fieldNames
	 */
	protected Field.Index[] fieldIndex = { NOT_ANALYZED_NO_NORMS,
			NOT_ANALYZED_NO_NORMS, NOT_ANALYZED_NO_NORMS,
			NOT_ANALYZED_NO_NORMS, NOT_ANALYZED_NO_NORMS,
			NOT_ANALYZED_NO_NORMS, NOT_ANALYZED_NO_NORMS,
			NOT_ANALYZED_NO_NORMS, ANALYZED, ANALYZED, ANALYZED };

	/**
	 * Lucene TermVector storage -- parallel array to fieldNames
	 */
	protected Field.TermVector[] fieldTermVector = { NO, NO, NO, NO, NO, NO,
			NO, NO, WITH_POSITIONS_OFFSETS, WITH_POSITIONS_OFFSETS,
			WITH_POSITIONS_OFFSETS };

	public StreamDocument() {
		fields = new HashMap<String, String>();
		sentences = new ArrayList<Sentence>();
	}

	public StreamDocument(StreamItem si, String dirName) {
		fields = new HashMap<String, String>();
		fields.put(FIELD_STREAM_ID, si.getStream_id());
		fields.put(FIELD_DOC_ID, si.getDoc_id());
		fields.put(FIELD_DIR_NAME, dirName);
		fields.put(FIELD_SOURCE, si.getSource());
		fields.put(FIELD_SCHOST, si.getSchost());
		// XXXX Tmp moved fields.put(FIELD_ZULU_TIMESTAMP, si.getStream_time() .getZulu_timestamp());
		fields.put(FIELD_EPOCH_TICKS,
				Double.toString(si.getStream_time().getEpoch_ticks()));
		// XXX fields.put(FIELD_TITLE, toUTF8(si.getTitle()));
		fields.put(FIELD_BODY, toUTF8(si.getBody()));
		List<Sentence> serif = si.getBody().getSentences().get("serif");
		if(serif != null) {
			sentences = serif;
		}




		// XXX fields.put(FIELD_ANCHOR, toUTF8(si.getAnchor()));
	}

	public void add(String key, String value) {
		fields.put(key, value);
	}

	public String get(String key) {
		return fields.get(key);
	}

	public Document toLuceneDocument() {
		Document d = new Document();
		for (int i = 0; i < fieldNames.length; i++) {
			String fieldName = fieldNames[i];
			String value = fields.get(fieldName);
			if (value != null) {
				// note: always store
				d.add(new Field(fieldName, value, Field.Store.YES,
						fieldIndex[i], fieldTermVector[i]));
			}
		}
		return d;
	}

	private static String toUTF8(ContentItem ci) {
		String encoding = ci.getEncoding();
		if (ci.getClean_visible() == null) 
			return null;

		if (ci.getClean_visible().length() > 0) {
			if (!StringUtils.isEmpty(encoding)
					&& !StringUtils.equalsIgnoreCase(encoding, "utf-8")) {
				return new String(ci.getClean_visible().getBytes(), Charset.forName("UTF-8"));
			}
			return new String(ci.getClean_visible());
		} else {
			return null;
		}
	}

	public List<Sentence> getSentences() {
		return sentences;
	}

	public boolean contains(String needle) {
		needle = needle.toLowerCase();

		// do matches case-insensitive
		String title = StringUtils.defaultString(StringUtils.lowerCase(fields
				.get(FIELD_TITLE)));
		if (title.contains(needle))
			return true;

		String body = StringUtils.defaultString(StringUtils.lowerCase(fields
				.get(FIELD_BODY)));
		if (body.contains(needle)) {
			return true;
		}
		String anchor = StringUtils.defaultString(StringUtils.lowerCase(fields
				.get(FIELD_ANCHOR)));
		if (anchor.contains(needle)) {
			return true;
		}
		return false;
	}

	public String getTextNormalized() {
		return StringUtils.defaultString(StringUtils.lowerCase(fields
				.get(FIELD_TITLE)))
				+ " "
				+ StringUtils.defaultString(StringUtils.lowerCase(fields
						.get(FIELD_BODY)))
				+ " "
				+ StringUtils.defaultString(StringUtils.lowerCase(fields
						.get(FIELD_ANCHOR)));
	}

	public void write(DataOutput out) throws IOException {
		for (Map.Entry<String, String> e : fields.entrySet()) {
			if (!StringUtils.isEmpty(e.getValue())) {
				Text.writeString(out, e.getKey());
				Text.writeString(out, e.getValue());
			}
		}
		Text.writeString(out, "EOD");

		out.writeInt(sentences.size());
		for (Sentence sent : sentences) {
			out.writeInt(sent.getTokensSize());
			for (Token tok : sent.getTokens()) {
				Text.writeString(out, tok.token);
				Text.writeString(out, tok.pos);
				Text.writeString(out, tok.dependency_path);
				if (tok.entity_type == null) {
					out.writeInt(-1);
				} else {
					out.writeInt(tok.entity_type.getValue());
				}
				if (tok.mention_type == null) {
					out.writeInt(-1);
				} else {
					out.writeInt(tok.mention_type.getValue());
				}
				out.writeInt(tok.token_num);
				out.writeInt(tok.parent_id);
				out.writeInt(tok.mention_id);
			}
		}
	}

	public void readFields(DataInput in) throws IOException {
		while (true) {
			String k = Text.readString(in);
			if (k.equals("EOD"))
				break;
			String v = Text.readString(in);
			fields.put(k, v);
		}

		int numSent = in.readInt();
		sentences = new ArrayList<Sentence>();
		for (int s = 0; s < numSent; s++) {
			Sentence tmpSent = new Sentence();
			int numTok = in.readInt();
			for (int t = 0; t < numTok; t++) {
				boolean ignoreToken = false;
				Token tmpTok = new Token();

				tmpTok.token = Text.readString(in);
				// If the token is in the filtered list we want to ignore it after it is deserialized
				if (NlpHelper.getTokenFilter().contains(tmpTok.token.toLowerCase())) {
					ignoreToken = true;
				}

				tmpTok.pos = Text.readString(in);
				tmpTok.dependency_path = Text.readString(in);
				tmpTok.entity_type = EntityType.findByValue(in.readInt());
				tmpTok.mention_type = MentionType.findByValue(in.readInt());
				tmpTok.token_num = in.readInt();
				tmpTok.parent_id = in.readInt();
				tmpTok.mention_id = in.readInt();

				if (!ignoreToken) {
					tmpSent.getTokens().add(tmpTok);
				}
			}
			sentences.add(tmpSent);
		}
	}

	public static StreamDocument read(DataInput in) throws IOException {
		StreamDocument doc = new StreamDocument();
		doc.readFields(in);
		return doc;
	}

	public String toString() {
		return "{" + "fields=" + fields + '}';
	}

}
