package no.ntnu.trec.kba;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import no.ntnu.trec.kba.bin.FeatureExtractorHadoop;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.log4j.Logger;

public class TemporalEntityData {
	private DateFormat inputDateFormat = new SimpleDateFormat("yyyy-MM-dd");
	private final static double BURST_THRESHOLD = 2.0;

	private int sumWPCountTraining;
	private int sumStreamCountTraining;

	private List<String> hours;
	private List<Integer> wpCounts;
	private List<Integer> wpCountsTraining;
	private List<Integer> streamCounts;
	private List<Integer> streamCountsTraining;
	private Map<String, Integer> hourIndex;

	public TemporalEntityData(Path timelineFile, FileSystem fs) throws ParseException {
		sumWPCountTraining = 0;
		sumStreamCountTraining = 0;
		
		hours = new ArrayList<String>();
		wpCounts = new ArrayList<Integer>();
		wpCountsTraining = new ArrayList<Integer>();
		streamCounts = new ArrayList<Integer>();
		streamCountsTraining = new ArrayList<Integer>();
		hourIndex = new HashMap<String, Integer>();

		int numHours = 0;
		
		try {
			BufferedReader input = new BufferedReader(new InputStreamReader(fs.open(timelineFile)));
			String line;
			
			String entityTrainingStop = input.readLine();
			
			Date endTrainingDate = inputDateFormat.parse(entityTrainingStop);
			boolean overTrainingPeriod = false;
			
			while ((line = input.readLine()) != null) {
				String[] split = line.split("\\s");
				if (split.length == 2) {
					String timestamp = split[0];
					int streamCount = new Integer(split[1]).intValue();
					hours.add(timestamp);
					hourIndex.put(timestamp, numHours);
					streamCounts.add(streamCount);
					// training period
					if (!overTrainingPeriod) {
						String t = timestamp.substring(0, timestamp.length()-3);
						Date now = inputDateFormat.parse(t);
						if (now.after(endTrainingDate)) {
							overTrainingPeriod = true;
						} else {
							streamCountsTraining.add(streamCount);
							sumStreamCountTraining += streamCount;
						}
					}
					numHours++;
				}
			}
			input.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private double getVolume(List<Integer> counts, String timestamp,
			int h) {
		int sum = 0;
		int num = 0;
		int idx = hourIndex.get(timestamp) - h;
		for (int i = 0; i < h; i++) {
			if (idx >= 0) {
				sum += counts.get(idx);
				num++;
			}
			idx++;
		}
		return (double) sum / num;
	}

	private Double getMoment(int moment, List<Double> set, double mean) {
		Double tmp = 0.d;
		for (Double sample : set) {
			tmp += Math.pow((sample-mean), moment);
		}

		return tmp/set.size();
	}

	private Double getMean(List<Double> set) {
		Double tmp = 0.d;
		for (Double entry : set) {
			tmp += entry;
		}

		return tmp/set.size();
	}

	private float getKurtosis(List<Double> set, double mean) {
		return (float)(getMoment(4, set, mean)/Math.pow(getMoment(2, set, mean), 2));
	}

	public double getStreamVolume(String timestamp, int h) {
		return getVolume(streamCounts, timestamp, h);
	}

	public double getStreamVolumeChange(String timestamp, int h) {
		return getStreamVolume(timestamp, h) / getAvgStreamVolume();
	}

	public List<Double> getSample(String timestamp, int h) {
		ArrayList<Double> sample = new ArrayList<Double>();
		int idx = hourIndex.get(timestamp) - h;
		for (int i = 0; i < h; i++) {
			if (idx >= 0) {
				sample.add((double)streamCounts.get(idx));
			}
			idx++;
		}

		return sample;
	}

	public double getStreamKurtosis(String timestamp, int h) {
		List<Double> sample = getSample(timestamp, h);
		Double mean = getMean(sample);

		double kurt = getKurtosis(sample, mean);

		if (Double.isNaN(kurt)) {
			return 0;
		}

		return kurt;
	}

	public double getTrend(String timestamp, int h) {
		List<Double> sample = getSample(timestamp, h);

		double xy = 0, x = 0, y = 0, x2 = 0;

		int x_val = 0;
		for (Double y_val : sample) {
			xy += x_val*y_val;
			x += x_val;
			y += y_val;
			x2 += Math.pow(x_val, 2);

			x_val++;
		}

		double trend = (sample.size()*xy-x*y)/(sample.size()*x2-Math.pow(x, 2));
		if (Double.isNaN(trend)) {
			return 0;
		}

		return trend;
	}

	public int isBurstStream(String timestamp, int h) {
		if (getStreamVolumeChange(timestamp, h) >= BURST_THRESHOLD)
			return 1;
		return 0;
	}

	public double getAvgStreamVolume() {
		// compute it only over the training period
		return (double) sumStreamCountTraining
				/ streamCountsTraining.size();
	}

}