package no.ntnu.trec.kba;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import no.ntnu.trec.kba.bin.FeatureExtractorHadoop;
import no.ntnu.utils.EntityNameResolver;
import no.ntnu.utils.MentionTextHelper;
import no.ntnu.utils.TextAnalysis;

import org.apache.commons.lang.StringUtils;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.LocatedFileStatus;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.RemoteIterator;
import org.apache.log4j.Logger;
import org.apache.lucene.analysis.Analyzer;

/**
 * Represents an input topic entity with id, name, and name variants. Related
 * entities and the textual representation associated with the entity (e.g., its
 * Wikipedia page) can be stored here, too.
 * 
 * NOTE: Entity name variants and related entities are analyzed using the
 * EntityAnalyzer (TextAnalysis.getAnalyzer()); text is analyzed using the
 * StandardAnalyzer (TextAnalysis.getStandardAnalyzer()).
 * 
 * @author Krisztian Balog <krisztian.balog@idi.ntnu.no>
 */
public class TopicEntity {

	private String urlname;

	private Set<String> nameVariants;

	private Set<String> nameVariantsAnalyzed;

	private Pattern nameVariantsPattern;

	private Set<String> relatedEntities;

	private Set<String> relatedEntitiesAnalyzed;

	private TemporalEntityData temporal;


	private Analyzer analyzer;

	private String text;

	private String[] textAnalyzed;

	private static final Logger log = Logger.getLogger(TopicEntity.class);


	public TopicEntity(String urlname) {
		analyzer = TextAnalysis.getAnalyzer();
		this.urlname = urlname;
		nameVariants = new HashSet<String>();
		nameVariantsAnalyzed = new HashSet<String>();
		languageModel = null;
		citations = new ArrayList<String[]>();
		wikiBursts = null;
		streamBursts = null;
		professionVector = null;
		// add the name itself to the variants
		// (it's important to set analyzer before this is called)
		addNameVariant(urlname);

		relatedEntities = new HashSet<String>();
		relatedEntitiesAnalyzed = new HashSet<String>();
		text = null;
	}

	public String getName() {
		return urlname;
	}

	public String[] getNameVariants() {
		return (String[]) nameVariants.toArray(new String[nameVariants.size()]);
	}

	public List<String[]> getNameVariantsAnalyzedAsList() {
		List<String[]> nva = new ArrayList<String[]>();
		for (String nv : nameVariantsAnalyzed) {
			nva.add(nv.split(" "));
		}
		return nva;
	}

	public Pattern getNameVariantsPattern() {
		return this.nameVariantsPattern;
	}

	public void addNameVariant(String name) {
		// deal with %XX
		if (name.contains("%")) {
			try {
				name = java.net.URLDecoder.decode(name, "UTF-8");
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			}
		}

		// the entity's WP page title is also used as a name variant
		// we have to clean that
		// Remove text between "(..)" and replace "_" with " "
		name = name.replaceAll("\\(.*\\)", "").replace("_", " ");

		// name variants are stored in a hashset, so we won't have duplicates
		nameVariants.add(name);

		// create analyzed version too
		// we store these as " " separated strings (not as String[]) so that
		// it's easier to deal with duplicates
		nameVariantsAnalyzed.add(StringUtils.join(
				TextAnalysis.analyze(analyzer, name), " "));

		rebuildNameVariantsPattern();
	}

	private void rebuildNameVariantsPattern() {
		this.nameVariantsPattern = MentionTextHelper.getPatternFromNameVariants(Arrays.asList(this.getNameVariants()));
	}

	/**
	 * Add name variants from a text file (one name variant per line)
	 * 
	 * @param filename
	 */
	public void addNameVariantsFromFile(Path path, FileSystem fs) {
		try {
			log.info("Reading in name variants for '" + getName() + "'");
			BufferedReader input = new BufferedReader(new InputStreamReader(fs.open(path)));

			String line = null;
			while ((line = input.readLine()) != null) {
					addNameVariant(line);
			}
			input.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void addRelatedEntity(String name) {
		// clean name
		try {
			name = java.net.URLDecoder.decode(name, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}
		// Remove text between "(..)" and replace "_" with " "
		name = name.replaceAll("\\(.*\\)", "").replace("_", " ");

		// name variants are stored in a hashset, so we won't have duplicates
		relatedEntities.add(name);

		// create analyzed version too
		// we store these as " " separated strings (not as String[]) so that
		// it's easier to deal with duplicates
		relatedEntitiesAnalyzed.add(StringUtils.join(
				TextAnalysis.analyze(analyzer, name), " "));
	}

	/**
	 * Add related entities from a text file (one entity per line)
	 * 
	 * @param filename
	 */
	public void addRelatedEntitiesFromFile(Path path, FileSystem fs) {
		try {
			log.info("Reading in related entity data for '" + getName() + "'");
			BufferedReader input = new BufferedReader(new InputStreamReader(fs.open(path)));
			String line = null;
			while ((line = input.readLine()) != null) {
				if (line.length() > 0)
					addRelatedEntity(line);
			}
			input.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void addTemporalDataFromfile(Path path, FileSystem fs) {
		try {
			log.info("Reading in timeline data for '" + getName() + "'");
			temporal = new TemporalEntityData(path, fs);
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}

	public Set<String> getRelatedEntities() {
		return relatedEntities;
	}

	public List<String[]> getRelatedEntitiesAnalyzed() {
		List<String[]> rea = new ArrayList<String[]>();
		for (String nv : relatedEntitiesAnalyzed) {
			rea.add(nv.split(" "));
		}
		return rea;
	}

	public String getText() {
		return text;
	}

	public String[] getTextAnalyzed() {
		return textAnalyzed;
	}

	public void setText(String text) {
		this.text = text;
		this.textAnalyzed = TextAnalysis.analyze(
				TextAnalysis.getStandardAnalyzer(), text);
	}

	/**
	 * Add text from a text file
	 * 
	 * @param filename
	 */
	public void addTextFromFile(Path path, FileSystem fs) {
		String text = "";
		try {
			log.info("Reading in entity profile for '" + getName() + "'");
			BufferedReader input = new BufferedReader(new InputStreamReader(fs.open(path)));
			String line = null;
			while ((line = input.readLine()) != null) {
				if (line.length() > 0)
					text = text.concat(line + "\n");
			}
			input.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		setText(text);
	}

	@Override
	public String toString() {
		String str = "'" + urlname + "'\n" + "Name variants:\n";
		for (String nv : nameVariants) {
			str += "  - '" + nv + "'\n";
		}
		str += "Analyzed name variants:\n";
		for (String nva : nameVariantsAnalyzed) {
			str += "  - '" + nva + "'\n";
		}
		return str;
	}

	public Analyzer getAnalyzer() {
		return analyzer;
	}

	public TemporalEntityData getTemporalData() {
		return temporal;
	}
}
