package no.ntnu.trec.kba.bin;

import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.utils.DocumentCleaner;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashSet;

/**
 * Perform basic document cleaning
 *
 * Specifically removes runs of whitespace and HTML tags
 * This also combines SequenceFiles for ease of use later
 *
 * Parameters:
 * -c: SequenceFile directory containing StreamDocuments
 * -o: Output path
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */

public class CleanDocuments extends Configured implements Tool {

	private static final Logger log = Logger.getLogger(CleanDocuments.class);

	public static class MyMapper extends
			Mapper<Text, StreamDocument, Text, StreamDocument> {

		private final Text keyout = new Text();

		@Override
		public void setup(Context context) throws IOException,
				InterruptedException {
			super.setup(context);
		}

		/**
		 * The key is the filename, the value is the StreamItem
		 */
		@Override
		public void map(Text key, StreamDocument value, Context context)
				throws IOException, InterruptedException {

			String streamId = value.get(StreamDocument.FIELD_STREAM_ID);

			keyout.set(streamId);
			String text = value.get(StreamDocument.FIELD_BODY);
			text = DocumentCleaner.clean(text);
			value.add(StreamDocument.FIELD_BODY, text);
			context.write(keyout, value);
		}
	}

	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new CleanDocuments(),
				args);
		System.exit(res);
	}

	public int run(String[] args) throws Exception {

		String corpusDir = null;
		String outputDir = null;

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-c".equals(args[i])) {
					corpusDir = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				return -1;
			}
		}

		if (corpusDir == null || outputDir == null) {
			return -1;
		}

		log.info("Tool name: " + this.getClass().getName());
		log.info(" - corpus_dir: " + corpusDir);
		log.info(" - output_dir: " + outputDir);

		Configuration conf = getConf();

		@SuppressWarnings("deprecation")
		Job job = new Job(conf, "Clean documents");
		job.setJarByClass(CleanDocuments.class);

		job.setInputFormatClass(SequenceFileInputFormat.class);
		job.setMapperClass(MyMapper.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(StreamDocument.class);
		job.setOutputFormatClass(SequenceFileOutputFormat.class);
		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(StreamDocument.class);
		job.setNumReduceTasks(1);

		FileInputFormat.addInputPath(job, new Path(corpusDir));
		FileSystem.get(conf).delete(new Path(outputDir), true);
		TextOutputFormat.setOutputPath(job, new Path(outputDir));

		return job.waitForCompletion(true) ? 0 : 1;

	}
}
