package no.ntnu.trec.kba.bin;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashSet;

import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.trec.kba.StreamItemWritable;
import no.ntnu.trec.kba.ThriftFileInputFormat;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

/**
 * Converts StreamCorpus Thrift files into SequenceFiles
 *
 * This also filters the corpus given a list of stream-ids
 *
 * Parameters:
 * -c: Path to StreamCorpus in thrift format
 * -s: Path to file containing a list of stream-ids to filter on (No filtering if not provided)
 * -o: Path to output
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */

public class ExtractDocuments extends Configured implements Tool {

	public static final String CONF_STREAM_IDS_FILE = "kba.stream_ids_file";


	private static final Logger log = Logger.getLogger(ExtractDocuments.class);

	public static class MyMapper extends
			Mapper<Text, StreamItemWritable, Text, StreamDocument> {

		private HashSet<String> streamIds;

		private final Text keyout = new Text();

		/**
		 * Load a set of stream_ids that we'd like to extract
		 * 
		 * @param context
		 * @throws IOException
		 */
		private void loadStreamIds(Context context) throws IOException {

			Configuration conf = context.getConfiguration();
			String fileName = conf.get(CONF_STREAM_IDS_FILE);

			if (fileName == null) {
				streamIds = null;
				return;
			}

			log.info("Loading stream_ids from " + fileName);

			streamIds = new HashSet<String>();

			try {
				FileSystem hdfs = FileSystem.get(conf);
				FSDataInputStream in = hdfs.open(new Path(fileName));
				BufferedReader br = new BufferedReader(
						new InputStreamReader(in));

				String line;
				while ((line = br.readLine()) != null) {
					if (line.length() > 0) {
						streamIds.add(line.split("\t")[0]);
					}
				}
				in.close();
			} catch (IOException e) {
				e.printStackTrace();
			} finally {
			}
			log.info(streamIds.size() + " stream_ids loaded");
		}

		@Override
		public void setup(Context context) throws IOException,
				InterruptedException {
			super.setup(context);
			loadStreamIds(context);
		}

		/**
		 * The key is the filename, the value is the StreamItem
		 */
		@Override
		public void map(Text key, StreamItemWritable value, Context context)
				throws IOException, InterruptedException {

			String streamId = value.getStream_id();
			
			// check if the stream item is on our 'wanted' list
			if (streamIds == null || streamIds.contains(streamId)) {
				String dirName = new Path(key.toString()).getParent().getName();

				// create StreamDocument from StreamItem
				StreamDocument doc = new StreamDocument(value, dirName);
				keyout.set(streamId);
				context.write(keyout, doc);
			}
		}
	}

	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new ExtractDocuments(),
				args);
		System.exit(res);
	}

	public int run(String[] args) throws Exception {

		String corpusDir = null;
		String streamIdsFile = null;
		String outputDir = null;

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-c".equals(args[i])) {
					corpusDir = args[++i];
				} else if ("-s".equals(args[i])) {
					streamIdsFile = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				}

			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				return -1;
			}
		}

		if (corpusDir == null || outputDir == null) {
			return -1;
		}

		log.info("Tool name: " + this.getClass().getName());
		log.info(" - corpus_dir: " + corpusDir);
		log.info(" - stream_ids_file: " + streamIdsFile);
		log.info(" - output_dir: " + outputDir);

		Configuration conf = getConf();
		if (streamIdsFile != null) {
			conf.set(CONF_STREAM_IDS_FILE, streamIdsFile);
		}

		@SuppressWarnings("deprecation")
		Job job = new Job(conf, "Document extractor");
		job.setJarByClass(ExtractDocuments.class);

		job.setInputFormatClass(ThriftFileInputFormat.class);
		job.setMapperClass(MyMapper.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(StreamDocument.class);
		job.setOutputFormatClass(SequenceFileOutputFormat.class);
		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(StreamDocument.class);

		job.setNumReduceTasks(25);

		FileInputFormat.addInputPath(job, new Path(corpusDir));
		FileSystem.get(conf).delete(new Path(outputDir), true);
		TextOutputFormat.setOutputPath(job, new Path(outputDir));

		return job.waitForCompletion(true) ? 0 : 1;

	}
}
