package no.ntnu.trec.kba.bin;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.trec.kba.StreamItemWritable;
import no.ntnu.trec.kba.ThriftFileInputFormat;
import no.ntnu.trec.kba.TopicEntity;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

/**
 * Extracts mentions of target entities from StreamCorpus
 *
 *
 * Parameters:
 * -c: Path to StreamCorpus in thrift format
 * -nv: Path to surface-forms
 * -o: Path to output
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */

public class ExtractMentions extends Configured implements Tool {

	public static final String CONF_TOPIC_FILE = "kba.topic_file";

	private static final Logger log = Logger.getLogger(ExtractMentions.class);

	public static class MyMapper extends
			Mapper<Text, StreamItemWritable, Text, Text> {
		private ArrayList<TopicEntity> topicEntities;
		private final Text keyout = new Text();

		@Override
		public void setup(Context context) throws IOException,
				InterruptedException {
			super.setup(context);
			loadTopicEntities(context);
		}

		private void loadTopicEntities(Context context) {
			Configuration conf = context.getConfiguration();
			String topicFile = conf.get(CONF_TOPIC_FILE);
			log.info("Loading topic entities from " + topicFile);
			try {
				HashMap<String, TopicEntity> entities = new HashMap<String, TopicEntity>();
				String[] topicEntityList = loadTextFile(topicFile, conf);
				for (int i = 0; i < topicEntityList.length; i++) {
					String[] split = topicEntityList[i].split("\\t+");
					String variant = split[0];
					String entityName = "https://kb.diffeo.com/resource/" + split[1];
					
					if (!entities.containsKey(entityName)) {
						entities.put(entityName, new TopicEntity(entityName));
					}
					entities.get(entityName).addNameVariant(variant);
				}
				topicEntities = new ArrayList<TopicEntity>(entities.values());
				/*for (TopicEntity entity : topicEntities) {
					log.info(entity.getName() + " loaded with "
							+ entity.getNameVariantsCount()
							+ " name variants");
				}
				*/
			} catch (Exception e) {
				log.error("Error while loading topic entities and name variants: " + e.getMessage());
				e.printStackTrace();
			}

			log.info(topicEntities.size() + " topic entities loaded");
		}

		/**
		 * The key is the filename, the value is the StreamItem
		 */
		@Override
		public void map(Text key, StreamItemWritable value, Context context)
				throws IOException, InterruptedException {

			String dirName = new Path(key.toString()).getParent().getName();

			// System.out.println("Dirname: "+dirName+" Value: "+value.getStream_id());

			// create StreamDocument from StreamItem
			StreamDocument doc = new StreamDocument(value, dirName);

			// for each topic entity, check if it is mentioned in the document
			for (TopicEntity entity : topicEntities) {
				// check all name variants of entity
				boolean mentioned = false;
				for (String nv : entity.getNameVariants()) {
					if (doc.contains(nv)) {
						mentioned = true;
						break;
					}
				}

				// if mentioned, pass <dirName, stream_id + urlname> to reducer
				if (mentioned) {
					keyout.set(dirName);
					context.write(keyout,
							new Text(doc.get(StreamDocument.FIELD_STREAM_ID)
									+ "\t" + entity.getName()));
				}
			}
		}
	}

	public static class MyReducer extends
			Reducer<Text, Text, NullWritable, Text> {

		public void setup(Context context) throws InterruptedException,
				IOException {
			super.setup(context);
		}

		@Override
		public void reduce(Text key, Iterable<Text> values, Context context)
				throws IOException, InterruptedException {
			for (Text t : values) {
				context.write(NullWritable.get(), t);
			}
			context.progress();
		}
	}

	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new ExtractMentions(),
				args);
		System.exit(res);
	}

	/**
	 * General method to read all lines of a text file either from local FS or
	 * from HDFS
	 * 
	 * @param fileName
	 * @return
	 */
	private static String[] loadTextFile(String fileName, Configuration conf) {
		List<String> lines = new ArrayList<String>();
		try {
			FileSystem hdfs = FileSystem.get(conf);
			FSDataInputStream in = hdfs.open(new Path(fileName));
			BufferedReader br = new BufferedReader(new InputStreamReader(in));

			String line;
			while ((line = br.readLine()) != null) {
				if (line.length() > 0)
					lines.add(line);
			}
			in.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return lines.toArray(new String[lines.size()]);
	}

	public int run(String[] args) throws Exception {

		String topicFile = null;
		String corpusDir = null;
		String outputDir = null;

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-nv".equals(args[i])) {
					topicFile = args[++i];
				} else if ("-c".equals(args[i])) {
					corpusDir = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				return -1;
			}
		}

		if (topicFile == null || corpusDir == null || outputDir == null) {
			return -1;
		}

		log.info("Tool name: " + this.getClass().getName());
		log.info(" - topic_file: " + topicFile);
		log.info(" - corpus_dir: " + corpusDir);
		log.info(" - output_dir: " + outputDir);

		Configuration conf = getConf();
		
		// Allow some failures in map splits
		// conf.set("mapreduce.map.failures.maxpercent","1");
		conf.set(CONF_TOPIC_FILE, topicFile);

		@SuppressWarnings("deprecation")
		Job job = new Job(conf);
		job.setJobName("KBA - Extract Mentions");
		job.setJarByClass(ExtractMentions.class);

		job.setInputFormatClass(ThriftFileInputFormat.class);
		job.setMapperClass(MyMapper.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(Text.class);
		job.setReducerClass(MyReducer.class);
		job.setOutputKeyClass(NullWritable.class);
		job.setOutputValueClass(Text.class);

		// single reduce task
		job.setNumReduceTasks(1);

		FileInputFormat.addInputPath(job, new Path(corpusDir));
		FileSystem.get(conf).delete(new Path(outputDir), true);
		TextOutputFormat.setOutputPath(job, new Path(outputDir));

		return job.waitForCompletion(true) ? 0 : 1;

	}
}
