package no.ntnu.trec.kba.bin;

import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.trec.kba.TextArrayWritable;
import no.ntnu.trec.kba.TopicEntity;
import no.ntnu.trec.kba.features.FeatureFactory;
import no.ntnu.trec.kba.features.FeatureGenerator;
import no.ntnu.utils.EntityNameResolver;
import no.ntnu.utils.HadoopArguments;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.MapFile;
import org.apache.hadoop.io.MapFile.Reader;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.concurrent.TimeUnit;

/**
 * Extract features using a specified feature extractor
 *
 * This variant does NOT use MapReduce and is therefore slower
 * Only use this if you cannot use the MapReduce variant
 * This only applies to the Clustering features and features which load word-vectors
 *
 * Parameters:
 * -i: SequenceFile directory containing StreamDocuments
 * -m: Path mentions MapFile
 * -f: Name of feature extractor to use
 * -o: Path to output
 * -F*: Feature extractor specific parameters
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */


public class FeatureExtractor extends Configured implements Tool {

	private static final Logger log = Logger.getLogger(FeatureExtractor.class);

	
	public static final String CONF_FEATURE_SET = "kba.feature_set";
	public static final String CONF_MENTIONS_DIR = "kba.mentions_dir";


	FeatureGenerator fg = null;
	private HashMap<String, TopicEntity> topicEntities = null;
	private MapFile.Reader mentionsReader = null;
	private SequenceFile.Reader corpusReader = null;
	private SequenceFile.Writer featureOutputWriter = null;

	Configuration conf = null;

		private TopicEntity getTopicEntity(String urlname) throws IOException {
			TopicEntity entity = null;
			FileSystem fs = FileSystem.get(conf);
			urlname = EntityNameResolver.removeNamePrefix(urlname);

			// we load each topic entity only once
			if (!topicEntities.containsKey(urlname)) {
				entity = new TopicEntity(urlname);
				
				// load name variants
				if (HadoopArguments.getArgFromConfig(conf, "Fnv") != null) {
					Path nvFile = new Path(HadoopArguments.getArgFromConfig(conf, "Fnv") + File.separator + urlname);
					if (fs.exists(nvFile))
						entity.addNameVariantsFromFile(nvFile, fs);
					else
						throw new IOException("Unable to open nv file: " + nvFile.toString());
				}
				
				// load related entities
				if (HadoopArguments.getArgFromConfig(conf, "Fr") != null) {
					Path rFile = new Path(HadoopArguments.getArgFromConfig(conf, "Fr") + File.separator + urlname);
					if (fs.exists(rFile))
						entity.addRelatedEntitiesFromFile(rFile, fs);
					else
						throw new IOException("Unable to open r file: " + rFile);
				}
				
				// load text representation of the entity
				if (HadoopArguments.getArgFromConfig(conf, "Fwp") != null) {
					Path wpFile = new Path(HadoopArguments.getArgFromConfig(conf, "Fwp") + File.separator + urlname);
					if (fs.exists(wpFile))
						entity.addTextFromFile(wpFile, fs);
					else
						throw new IOException("Unable to open wp file: " + wpFile);
				}
				
				// load temporal data
				if (HadoopArguments.getArgFromConfig(conf, "Ft") != null) {
					Path tFile = new Path(HadoopArguments.getArgFromConfig(conf, "Ft") + File.separator + urlname);
					if (fs.exists(tFile))
						entity.addTemporalDataFromfile(tFile, fs);
					else
						throw new IOException("Unable to open t file: " + tFile);
				}
				
				
				log.debug("Topic entity '" + urlname + "' loaded:\n"
						+ entity.toString());

				// cache it
				topicEntities.put(urlname, entity);
			} else {
				entity = topicEntities.get(urlname);
			}
			return entity;
		}

		public void process() throws IOException {
			Text key = new Text();
			StreamDocument doc = new StreamDocument();

			long lastCheckpoint = System.currentTimeMillis();
			int processed = 0;
			while (corpusReader.next(key, doc)) {
				TextArrayWritable mentions = new TextArrayWritable();

				if (mentionsReader.get(key, mentions) == null) {
					return;
				}

				for (Writable mention : mentions.get()) {
					TopicEntity ent = getTopicEntity(mention.toString());
					String result = fg.getFeatureVector(doc, ent);

					featureOutputWriter.append(new Text(key + "\t" + mention), new Text(result));
				}

				processed++;
				if (processed % 10000 == 0) {
					log.info("Checkpoint: " + processed + " docs done");
					log.info("Time to process last 10k docs: " + TimeUnit.MILLISECONDS.toSeconds(System.currentTimeMillis()-lastCheckpoint) + "s");
					lastCheckpoint = System.currentTimeMillis();
				}
			}
			featureOutputWriter.close();
			corpusReader.close();
		}

	public int run(String[] args) throws Exception {
		String inputDir = null;
		String mentionsFile = null;
		String featureSet = null;
		String outputDir = null;

		conf = getConf();


		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-i".equals(args[i])) {
					inputDir = args[++i];
				} else if ("-m".equals(args[i])) {
					mentionsFile = args[++i];
				} else if ("-f".equals(args[i])) {
					featureSet = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				return -1;
			}
		}

		HadoopArguments.addArgsToConfig(conf, args);

		if (inputDir == null || mentionsFile == null || featureSet == null
				|| outputDir == null) {
			return -1;
		}
		
		conf.set(CONF_MENTIONS_DIR, mentionsFile);
		conf.set(CONF_FEATURE_SET, featureSet);
		
		FileSystem.get(conf).delete(new Path(outputDir), true);

		topicEntities = new HashMap<String, TopicEntity>();

		mentionsReader = new Reader(new Path(mentionsFile), conf);
		corpusReader = new SequenceFile.Reader(conf, SequenceFile.Reader.file(new Path(inputDir)));
		featureOutputWriter = SequenceFile.createWriter(conf,
				SequenceFile.Writer.file(new Path(outputDir)),
				SequenceFile.Writer.keyClass(Text.class),
				SequenceFile.Writer.valueClass(Text.class),
				SequenceFile.Writer.compression(SequenceFile.CompressionType.NONE));
		// instantiate the particular feature extractor
		fg = FeatureFactory.getFeatureGenerator(featureSet, conf);
		if (fg == null) {
			log.error("Unknown feature set '" + featureSet + "'");
			System.exit(-1);
		}

		process();

		return 0;

	}

	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new FeatureExtractor(),
				args);
		System.exit(res);
	}
}
