package no.ntnu.trec.kba.bin;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;

import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.trec.kba.TextArrayWritable;
import no.ntnu.trec.kba.TopicEntity;
import no.ntnu.trec.kba.features.FeatureFactory;
import no.ntnu.trec.kba.features.FeatureGenerator;
import no.ntnu.utils.EntityNameResolver;

import no.ntnu.utils.HadoopArguments;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.MapFile.Reader;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;


/**
 * Extract features using a specified feature extractor
 *
 * Do NOT use this for feature extractor that either
 * (1) use extreme amounts of memory (ie. loads word-vectors)
 * (2) Require documents to be processed in-order (Clustering features)
 *
 * Parameters:
 * -i: SequenceFile directory containing StreamDocuments
 * -m: Path mentions MapFile
 * -f: Name of feature extractor to use
 * -o: Path to output
 * -F*: Feature extractor specific parameters
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */

public class FeatureExtractorHadoop extends Configured implements Tool {

	private static final Logger log = Logger.getLogger(FeatureExtractorHadoop.class);

	
	public static final String CONF_FEATURE_SET = "kba.feature_set";
	public static final String CONF_MENTIONS_DIR = "kba.mentions_dir";
	public static final String CONF_PREFIX = "kba.prefix";

	
	public static class MyMapper extends
	Mapper<Text, StreamDocument, Text, Text>  {
		
		FeatureGenerator fg = null; 
		private HashMap<String, TopicEntity> topicEntities = null;
		private HashMap<String, String> featureProps = null;
		private Reader reader = null;
		private String prefix = null;


		/**
		 * Return the TopicEntity object for a given urlname.
		 * (Takes care of the caching an loading.) 
		 * 
		 * @param urlname
		 * @return
		 * @throws IOException 
		 */
		private TopicEntity getTopicEntity(String urlname, Context context) throws IOException {
			TopicEntity entity = null;
			FileSystem fs = FileSystem.get(context.getConfiguration());
			Configuration config = context.getConfiguration();

			// we load each topic entity only once
			urlname = EntityNameResolver.removeNamePrefix(urlname);
			if (!topicEntities.containsKey(urlname)) {
				entity = new TopicEntity(urlname);
				
				// load name variants
				if (HadoopArguments.getArgFromConfig(config, "Fnv") != null) {
					Path nvFile = new Path(HadoopArguments.getArgFromConfig(config, "Fnv") + File.separator + urlname);
					if (fs.exists(nvFile))
						entity.addNameVariantsFromFile(nvFile, fs);
					else
						throw new IOException("Unable to open nv file: " + nvFile.toString());
				}
				
				// load related entities
				if (HadoopArguments.getArgFromConfig(config, "Fr") != null) {
					Path rFile = new Path(HadoopArguments.getArgFromConfig(config, "Fr") + File.separator + urlname);
					if (fs.exists(rFile))
						entity.addRelatedEntitiesFromFile(rFile, fs);
					else
						throw new IOException("Unable to open r file: " + rFile);
				}
				
				// load text representation of the entity
				if (HadoopArguments.getArgFromConfig(config, "Fwp") != null) {
					Path wpFile = new Path(HadoopArguments.getArgFromConfig(config, "Fwp") + File.separator + urlname);
					if (fs.exists(wpFile))
						entity.addTextFromFile(wpFile, fs);
					else
						throw new IOException("Unable to open wp file: " + wpFile);
				}
				
				// load temporal data
				if (HadoopArguments.getArgFromConfig(config, "Ft") != null) {
					Path tFile = new Path(HadoopArguments.getArgFromConfig(config, "Ft") + File.separator + urlname);
					if (fs.exists(tFile))
						entity.addTemporalDataFromfile(tFile, fs);
					else
						throw new IOException("Unable to open t file: " + tFile);
				}
				
				
				log.debug("Topic entity '" + urlname + "' loaded:\n"
						+ entity.toString());

				// cache it
				topicEntities.put(urlname, entity);
			} else {
				entity = topicEntities.get(urlname);
			}
			return entity;
		}
		
		@Override
		public void setup(Context context) throws IOException,
		InterruptedException {
			Configuration conf = context.getConfiguration();
			String mentionsDir = conf.get(CONF_MENTIONS_DIR);
			String featureSet = conf.get(CONF_FEATURE_SET);
			prefix = conf.get(CONF_PREFIX);
			topicEntities = new HashMap<String, TopicEntity>();
			
			reader = new Reader(new Path(mentionsDir), conf);

			// instantiate the particular feature extractor
			fg = FeatureFactory.getFeatureGenerator(featureSet, conf);
			if (fg == null) {
				log.error("Unknown feature set '" + featureSet + "'");
				System.exit(-1);
			}
			
			super.setup(context);


		}
		
		@Override
		public void map(Text key, StreamDocument value, Context context)
			throws IOException, InterruptedException {
			TextArrayWritable mentions = new TextArrayWritable();
			
			if (reader.get(key, mentions) == null) {
				return;
			}
			
			for (Writable mention : mentions.get()) {
				TopicEntity ent = getTopicEntity(mention.toString(), context);
				String result = fg.getFeatureVector(value, ent);
				context.write(new Text(key + "\t" + mention), new Text(result));
			}
		}
	}

	public int run(String[] args) throws Exception {
		String inputDir = null;
		String mentionsFile = null;
		String featureSet = null;
		String outputDir = null;

		Configuration conf = getConf();


		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-i".equals(args[i])) {
					inputDir = args[++i];
				} else if ("-m".equals(args[i])) {
					mentionsFile = args[++i];
				} else if ("-f".equals(args[i])) {
					featureSet = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				} else if ("-p".equals(args[i])) {
					conf.set(CONF_PREFIX, args[++i]);
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				return -1;
			}
		}

		HadoopArguments.addArgsToConfig(conf, args);

		if (inputDir == null || mentionsFile == null || featureSet == null
				|| outputDir == null) {
			return -1;
		}
		
		conf.set(CONF_MENTIONS_DIR, mentionsFile);
		conf.set(CONF_FEATURE_SET, featureSet);

		Job job = new Job(conf, "KBA - Feature Extraction - " + featureSet);
		job.setJarByClass(FeatureExtractorHadoop.class);

		job.setInputFormatClass(SequenceFileInputFormat.class);
		job.setMapperClass(MyMapper.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(Text.class);
		job.setOutputFormatClass(SequenceFileOutputFormat.class);
		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(Text.class);
		job.setNumReduceTasks(1);
		
		FileInputFormat.addInputPath(job, new Path(inputDir));
		FileSystem.get(conf).delete(new Path(outputDir), true);
		SequenceFileOutputFormat.setOutputPath(job, new Path(outputDir));

		return job.waitForCompletion(true) ? 0 : 1;
	}

	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new FeatureExtractorHadoop(),
				args);
		System.exit(res);
	}
}
