package no.ntnu.trec.kba.bin;

import no.ntnu.utils.EntityNameResolver;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.util.HashMap;
import java.util.regex.Pattern;

/**
 * Combines feature output files
 *
 * This is used for combining the feature output of several feature extractors
 *
 * Parameters:
 * -i: SequenceFile directory containing StreamDocuments
 * -o: Path to output
 * -e: If provided a regular expression then excludes features whos name match the pattern (Optional)
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */

public class FeatureOutputCombiner extends Configured implements Tool {
private static final Logger log = Logger.getLogger(ExtractDocuments.class);

	public static final String CONF_FILTER_FILE = "kba.filter_file";
	
	public static class MyMapper extends
	Mapper<Text, Text, Text, Text>  {
		@Override
		public void map(Text key, Text value, Context context)
			throws IOException, InterruptedException {

			context.write(key, value);
				
		}
	}

	public static class MyReducer extends
	Reducer<Text, Text, Text, Text> {
		Pattern excludeFilter = null;

		@Override
		public void setup(Reducer.Context context) throws IOException,
				InterruptedException {
			String filterString = context.getConfiguration().get(CONF_FILTER_FILE);

			if (filterString != null) {
				excludeFilter = Pattern.compile(filterString);
			}
		}

		@Override
		public void reduce(Text key, Iterable<Text> values, Context context) throws IOException, InterruptedException {
			HashMap<String, String> map = new HashMap<String, String>();
			for (Text val : values) {
				String[] split = val.toString().split("\t");
				for (String pair : split) {
					String[] keyVal = pair.split(":");
					if (excludeFilter == null || !excludeFilter.matcher(keyVal[0]).matches()) {
						map.put(keyVal[0], keyVal[1]);
					}
				}
			}

			String outVal = "";
			boolean first = true;
			for (String feat : map.keySet()) {
				if (first) {
					first = false;
				} else {
					outVal += "\t";
				}
				String val = map.get(feat);
				outVal += feat + ":" + val;
			}
			
			context.write(key, new Text(outVal));
		}
	}
	
	public int run(String[] args) throws Exception {

		String outputDir = null;
		String inputFile = null;
		String filterFile = null;

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-i".equals(args[i])) {
					inputFile = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				} else if ("-e".equals(args[i])) {
					filterFile = args[++i];
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				return -1;
			}
		}

		if (inputFile == null || outputDir == null) {
			return -1;
		}

		log.info("Tool name: " + this.getClass().getName());
		log.info(" - input_file: " + inputFile);
		log.info(" - output_dir: " + outputDir);

		Configuration conf = getConf();
		if (filterFile != null) {
			conf.set(CONF_FILTER_FILE, filterFile);
		}

		@SuppressWarnings("deprecation")
		Job job = new Job(conf, "KBA - Combine feature output");
		job.setJarByClass(MentionsToHadoopMapFile.class);

		job.setInputFormatClass(SequenceFileInputFormat.class);
		job.setMapperClass(MyMapper.class);
		job.setReducerClass(MyReducer.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(Text.class);
		job.setOutputFormatClass(SequenceFileOutputFormat.class);
		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(Text.class);

		SequenceFileInputFormat.addInputPath(job, new Path(inputFile));
		FileSystem.get(conf).delete(new Path(outputDir), true);
		TextOutputFormat.setOutputPath(job, new Path(outputDir));

		return job.waitForCompletion(true) ? 0 : 1;

	}
	
	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new FeatureOutputCombiner(),
				args);
		System.exit(res);
	}
}
