package no.ntnu.trec.kba.bin;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

/**
 * Generate arff files for classification
 * 
 * @author Krisztian Balog <krisztian.balog@idi.ntnu.no>
 */
public class GenerateArffFile {

	private static final Logger log = Logger.getLogger(GenerateArffFile.class);

	private List<String> featureTypes;
	private Map<String, List<String>> featureValues;

	GenerateArffFile() {
		featureTypes = new ArrayList<String>();
		featureValues = new HashMap<String, List<String>>();
	}

	private void loadFeatureFile(String filename) {
		log.info("Loading features from " + filename);
		try {
			BufferedReader br = new BufferedReader(new FileReader(filename));
			String line;
			int featureNum = 0;
			while ((line = br.readLine()) != null) {
				if (StringUtils.isBlank(line))
					continue;
				// attribute line
				if (line.startsWith("@")) {
					featureTypes.add(line);
					featureNum++;
				}
				// feature line
				else {
					String[] tmp = line.split("\\s+");
					// check if the number of features match
					if (tmp.length != featureNum + 2) {
						log.error("Missing features in\n  file: " + filename
								+ "\n  line: " + line);
						continue;
					}
					// key is streamId-urlname
					String key = tmp[0] + "-" + tmp[1];
					List<String> value;
					if (featureValues.containsKey(key))
						value = featureValues.get(key);
					else
						value = new ArrayList<String>();
					for (int i = 2; i < tmp.length; i++) {
						value.add(tmp[i]);
					}
					featureValues.put(key, value);
				}
			}
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void run(String annotationsFile, String mentionsFile,
			List<String> featureFiles, Set<Integer> positiveClass,
			Set<Integer> negativeClass, String outputFileTraining,
			String outputFileTesting) {

		// load pre-computed features
		for (String filename : featureFiles) {
			loadFeatureFile(filename);
		}

		generateTrainingArffFile(annotationsFile, positiveClass, negativeClass,
				outputFileTraining);
		generateTestingArffFile(mentionsFile, outputFileTesting);
	}
	
	private String getArffHeader() {
		String output = "@RELATION trec-kba\n";
		for (int i = 0; i < featureTypes.size(); i++) {
			output += featureTypes.get(i) + "\n";
		}
		output += "@ATTRIBUTE class {0,1}\n\n";
		output  += "@DATA\n";
		return output;
	}
	
	private String getArffFeatureVector(String key, String targetClass) {
		
		if (!featureValues.containsKey(key)) {
			log.error("Missing feature values for " + key + "\t" + targetClass);
			return null;
		}				
		
		String output = "";
		List<String> fv = featureValues.get(key);
		for (int i = 0; i < fv.size(); i++) {
			output += fv.get(i) + ",";
		}
		output += targetClass;
		
		return output;
	}

	private void generateTestingArffFile(String mentionsFile,
			String outputFileTesting) {
		log.info("Generating testing arff file based on " + mentionsFile);
		try {
			Writer output = new BufferedWriter(new FileWriter(
					outputFileTesting));
			output.write(getArffHeader());
			// write stream_id - urlname pairs to .ids file
			Writer output2 = new BufferedWriter(new FileWriter(
					outputFileTesting + ".ids"));

			// write feature vectors
			BufferedReader br = new BufferedReader(new FileReader(
					mentionsFile));
			String line;
			while ((line = br.readLine()) != null) {
				String[] tmp = line.split("\\s+");
				if (tmp.length != 2)
					continue;

				String streamId = tmp[0];
				String urlname = tmp[1];
				// key is streamId-urlname
				String key = streamId + "-" + urlname;
				String targetClass = "?";
				String fv = getArffFeatureVector(key, targetClass);
				if (fv != null) {
					output.write(fv + "\n");
					output2.write(streamId + "\t" + urlname + "\n");
				}
			}
			br.close();
			output.close();
			output2.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void generateTrainingArffFile(String annotationsFile,
			Set<Integer> positiveClass, Set<Integer> negativeClass,
			String outputFileTraining) {

		log.info("Generating training arff file based on " + annotationsFile);
		try {
			Writer output = new BufferedWriter(new FileWriter(
					outputFileTraining));
			output.write(getArffHeader());

			// write feature vectors
			BufferedReader br = new BufferedReader(new FileReader(
					annotationsFile));
			String line;
			while ((line = br.readLine()) != null) {
				if (line.startsWith("#") || StringUtils.isBlank(line))
					continue;

				String[] tmp = line.split("\\s+");
				if (tmp.length != 7)
					continue;

				// key is streamId-urlname
				String key = tmp[2] + "-" + tmp[3];
				Integer rel = new Integer(tmp[5]);
				Integer mention = new Integer(tmp[6]);

				// consider only explicit mentions
				if (mention != 1)
					continue;

				// consider only relevance levels that correspond to either the
				// positive or negative class
				if (!positiveClass.contains(rel) && !negativeClass.contains(rel))
					continue;
				
				String targetClass = positiveClass.contains(rel) ? "1" : "0";
				String fv = getArffFeatureVector(key, targetClass);
				if (fv != null)
					output.write(fv + "\n");
			}
			br.close();
			output.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	static void printUsage() {
		System.out
				.println("Usage: "
						+ GenerateArffFile.class.getName()
						+ " -a annotations_file -m mentions_file -f feature_file \n"
						+ " -o1 output_training -o2 output_testing [-pc positive_class] [-nc negative_class] \n\n"
						+ "   annotations_file : Annotations file (training data) \n"
						+ "   mentions_file    : List of stream_id-urlname pairs (testing data) \n"
						+ "   feature_file     : Precomputed feature values to be used \n"
						+ "                        File must contain both training and testing periods! \n"
						+ "                        Use \"-f feature_file\" multiple times for \n"
						+ "                        combining multiple feature sets \n"
						+ "   output_training : Output file for training data \n"
						+ "   output_testing  : Output file for testing data \n"
						+ "   positive_class  : Comma separated list of relevance levels that count \n"
						+ "                       as positive instances (default: 1,2) \n"
						+ "   negative_class  : Comma separated list of relevance levels that count \n"
						+ "                       as negative instances (default: -1,0) \n\n"
						+ "Example usage: java -jar trec-kba.jar "
						+ GenerateArffFile.class.getName() + " "
						+ "-i trec-kba/extracted-documents/* "
						+ "-m trec-kba/extract-mentions/dbpedia-2011 "
						+ "-f xxx" + "-o trec-kba/features/xxx \n\n");
		System.exit(-1);
	}

	static Set<Integer> parseIntList(String str) {
		Set<Integer> s = new HashSet<Integer>();
		String[] tmp = str.split(",");
		for (int i = 0; i < tmp.length; i++) {
			s.add(new Integer(tmp[i]));
		}
		return s;
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {

		String annotationsFile = null;
		String mentionsFile = null;
		List<String> featureFiles = new ArrayList<String>();
		String outputFileTraining = null;
		String outputFileTesting = null;
		Set<Integer> positiveClass = new HashSet<Integer>();
		Set<Integer> negativeClass = new HashSet<Integer>();

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-a".equals(args[i])) {
					annotationsFile = args[++i];
				} else if ("-m".equals(args[i])) {
					mentionsFile = args[++i];
				} else if ("-f".equals(args[i])) {
					featureFiles.add(args[++i]);
				} else if ("-o1".equals(args[i])) {
					outputFileTraining = args[++i];
				} else if ("-o2".equals(args[i])) {
					outputFileTesting = args[++i];
				} else if ("-pc".equals(args[i])) {
					positiveClass = parseIntList(args[++i]);
				} else if ("-nc".equals(args[i])) {
					negativeClass = parseIntList(args[++i]);
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				printUsage();
			}
		}

		if (annotationsFile == null || mentionsFile == null
				|| featureFiles == null || outputFileTraining == null
				|| outputFileTesting == null) {
			printUsage();
		}

		// use default pos class: R and C
		if (positiveClass.size() == 0) {
			positiveClass.add(1); // R
			positiveClass.add(2); // C
		}
		// use default neg class: G and N
		if (negativeClass.size() == 0) {
			negativeClass.add(-1); // G
			negativeClass.add(0); // N
		}

		GenerateArffFile gaf = new GenerateArffFile();
		gaf.run(annotationsFile, mentionsFile, featureFiles, positiveClass,
				negativeClass, outputFileTraining, outputFileTesting);
	}

}
