package no.ntnu.trec.kba.bin;

import java.io.IOException;
import java.util.HashSet;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.KeyValueTextInputFormat;
import org.apache.hadoop.mapreduce.lib.output.MapFileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

import no.ntnu.trec.kba.TextArrayWritable;


/**
 * Convert file containing mentions into Hadoop MapFile for fast lookup of mentions
 *
 * Parameters:
 * -i: Path to file containing raw mentions
 * -o: Path to output
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */

public class MentionsToHadoopMapFile extends Configured implements Tool {
	private static final Logger log = Logger.getLogger(ExtractDocuments.class);

	
	public static class MyMapper extends
	Mapper<Text, Text, Text, Text>  {
		@Override
		public void map(Text key, Text value, Context context)
			throws IOException, InterruptedException {
			
			String s = value.toString();
			Text t = new Text(s);
			context.write(key, t);
				
		}
	}

	public static class MyReducer extends
	Reducer<Text, Text, Text, TextArrayWritable> {
		@Override
		public void reduce(Text key, Iterable<Text> values, Context context) throws IOException, InterruptedException {
			
			HashSet<Text> content = new HashSet<Text>();
			for (Text val : values) {
				content.add(new Text(val));
			}
			Text[] arr = new Text[content.size()];
			arr = content.toArray(arr);
			TextArrayWritable newValue = new TextArrayWritable();
			newValue.set(arr);
			
			context.write(key, newValue);
		}
	}
	
	public int run(String[] args) throws Exception {

		String outputDir = null;
		String inputFile = null;

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-i".equals(args[i])) {
					inputFile = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				return -1;
			}
		}

		if (inputFile == null || outputDir == null) {
			return -1;
		}

		log.info("Tool name: " + this.getClass().getName());
		log.info(" - input_file: " + inputFile);
		log.info(" - output_dir: " + outputDir);

		Configuration conf = getConf();

		@SuppressWarnings("deprecation")
		Job job = new Job(conf, "KBA - Mentions to MapFile");
		job.setJarByClass(MentionsToHadoopMapFile.class);

		job.setInputFormatClass(KeyValueTextInputFormat.class);
		job.setMapperClass(MyMapper.class);
		job.setReducerClass(MyReducer.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(Text.class);
		job.setOutputFormatClass(MapFileOutputFormat.class);
		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(TextArrayWritable.class);
		job.setNumReduceTasks(1);

		FileInputFormat.addInputPath(job, new Path(inputFile));
		FileSystem.get(conf).delete(new Path(outputDir), true);
		TextOutputFormat.setOutputPath(job, new Path(outputDir));

		return job.waitForCompletion(true) ? 0 : 1;

	}
	
	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new MentionsToHadoopMapFile(),
				args);
		System.exit(res);
	}
}
