package no.ntnu.trec.kba.bin;

import java.io.IOException;
import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.trec.kba.TextArrayWritable;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.MapFile.Reader;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.mapreduce.lib.reduce.IntSumReducer;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;

/**
 *
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */


/**
 * Extracts the number of mentions of target entities in hourly batches
 *
 * Used for temporal features that depend on a time-series of mentions of each entity
 *
 * Parameters:
 * -i: SequenceFile directory containing StreamDocuments
 * -m: Path mentions MapFile
 * -o: Path to output
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */

public class TemporalMentionsExtractor extends Configured implements Tool {

	private static final Logger log = Logger.getLogger(TemporalMentionsExtractor.class);

	
	public static final String CONF_MENTIONS_DIR = "kba.mentions_dir";

	
	public static class MyMapper extends
	Mapper<Text, StreamDocument, Text, IntWritable>  {
		
		private Reader reader = null;
		
		@Override
		public void setup(Context context) throws IOException,
		InterruptedException {
			Configuration conf = context.getConfiguration();
			String mentionsDir = conf.get(CONF_MENTIONS_DIR);
			
			reader = new Reader(new Path(mentionsDir), conf);
			
			
			super.setup(context);
		}
		
		@Override
		public void map(Text key, StreamDocument value, Context context)
			throws IOException, InterruptedException {
			TextArrayWritable mentions = new TextArrayWritable();
			if (reader.get(key, mentions) == null) {
				throw new IOException(key + " has no mentions, this should not be possible.");
			}
			
			for (Writable mention : mentions.get()) {
				context.write(new Text(mention + "\t" + value.get(StreamDocument.FIELD_DIR_NAME)), new IntWritable(1));
			}
		}
	}

	public int run(String[] args) throws Exception {
		String inputDir = null;
		String mentionsFile = null;
		String outputDir = null;
		
		Configuration conf = getConf();


		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-i".equals(args[i])) {
					inputDir = args[++i];
				} else if ("-m".equals(args[i])) {
					mentionsFile = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
				return -1;
			}
		}

		if (inputDir == null || mentionsFile == null || outputDir == null) {
			return -1;
		}
		
		conf.set(CONF_MENTIONS_DIR, mentionsFile);


		Job job = new Job(conf, "Extract temporal stream data");
		job.setJarByClass(TemporalMentionsExtractor.class);

		job.setInputFormatClass(SequenceFileInputFormat.class);
		job.setMapperClass(MyMapper.class);
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(IntWritable.class);
		job.setReducerClass(IntSumReducer.class);
		job.setOutputFormatClass(TextOutputFormat.class);
		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(IntWritable.class);
		
		FileInputFormat.addInputPath(job, new Path(inputDir));
		FileSystem.get(conf).delete(new Path(outputDir), true);
		TextOutputFormat.setOutputPath(job, new Path(outputDir));

		return job.waitForCompletion(true) ? 0 : 1;
	}

	public static void main(String[] args) throws Exception {
		int res = ToolRunner.run(new Configuration(), new TemporalMentionsExtractor(),
				args);
		System.exit(res);
	}
}
