package no.ntnu.trec.kba.bin;

import edu.stanford.nlp.ie.util.RelationTriple;
import edu.stanford.nlp.process.Morphology;
import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.trec.kba.TextArrayWritable;
import no.ntnu.trec.kba.TopicEntity;
import no.ntnu.utils.*;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.MapFile;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.log4j.Logger;
import streamcorpus.Sentence;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;
import java.util.zip.GZIPInputStream;


/**
 * Extracts vocabulary of terms from training documents
 *
 * Used for generating vocabulary for features based on VSM
 * Remeber to run this only on training documents, or you will get a very large vocabulary
 *
 * Parameters:
 * -i: SequenceFile directory containing StreamDocuments
 * -m: Path to mentions MapFile
 * -nv: Path to surface-forms
 * -synsets: True if extracting vocabulart for hypernym augmented VSM, false for just verbs
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */

public class VocabularyExtractor extends Configured implements Tool {


    private static final Logger log = Logger.getLogger(no.ntnu.trec.kba.bin.VocabularyExtractor.class);

    public static final String CONF_MENTIONS_DIR = "kba.mentions_dir";
    public static final String CONF_VARIANT_DIR = "kba.variants_dir";
    public static final String CONF_USE_FULLTEXT = "kba.fulltext";


    public static class MyMapper extends
            Mapper<Text, StreamDocument, Text, NullWritable> {

        private MapFile.Reader reader = null;
        private HashMap<String, ArrayList<String>> nameVariants = null;
        private HashMap<String, TopicEntity> entities = null;
        private String nvDir = null;
        private boolean useSynsets = false;

        private TopicEntity getTopicEntity(String urlname, Context context) throws IOException {
            TopicEntity entity = null;
            FileSystem fs = FileSystem.get(context.getConfiguration());
            // we load each topic entity only once
            if (!entities.containsKey(urlname)) {
                urlname = EntityNameResolver.removeNamePrefix(urlname);
                entity = new TopicEntity(urlname);

                Path nvFile = new Path(nvDir + File.separator + urlname);
                if (fs.exists(nvFile))
                    entity.addNameVariantsFromFile(nvFile, fs);
                else
                    throw new IOException("Unable to open nv file: " + nvFile.toString());

                entities.put(urlname, entity);
            }

            return entities.get(urlname);
        }

        @Override
        public void setup(Context context) throws IOException,
                InterruptedException {
            Configuration conf = context.getConfiguration();
            nvDir = conf.get(CONF_VARIANT_DIR);
            useSynsets = Boolean.valueOf(conf.get(CONF_USE_FULLTEXT));
            reader = new MapFile.Reader(new Path(conf.get(CONF_MENTIONS_DIR)), conf);
            nameVariants = new HashMap<String, ArrayList<String>>();
            entities = new HashMap<String, TopicEntity>();

            super.setup(context);
        }

        @Override
        public void map(Text key, StreamDocument value, Context context)
                throws IOException, InterruptedException {
            TextArrayWritable mentions = new TextArrayWritable();

            if (reader.get(key, mentions) == null) {
                return;
            }

            HashSet<String> verbSet = new HashSet<String>();

            for (Writable mention : mentions.get()) {
                TopicEntity me = getTopicEntity(mention.toString(), context);

                // Get only the sentences that mention the entity
                List<Sentence> sentences = value.getSentences();
                sentences = NlpHelper.getMentioningSentences(sentences, me.getNameVariantsPattern());

                for (Sentence sent : sentences) {

                    List<RelationTriple> triples = NlpHelper.getMentioningTriples(sent, me.getNameVariantsPattern());

                    Map<String, String> verbs = NlpHelper.verbsAndTagsFromTriples(triples);

                    for (Map.Entry<String, String> entry : verbs.entrySet()) {
                        if (!useSynsets) {
                            verbSet.add(Morphology.lemmaStatic(entry.getKey(), entry.getValue()));
                        } else {
                            for (String hypernymOffset : WordnetHelper.getVerbHypernyms(entry.getKey(), entry.getValue())) {
                                verbSet.add(hypernymOffset);
                            }

                        }
                    }
                }
                for (String verb : verbSet) {
                    context.write(new Text(verb), NullWritable.get());
                }
            }
        }
    }

    public static class MyReducer extends
            Reducer<Text, NullWritable, Text, NullWritable> {

        @Override
        protected void reduce(Text key, Iterable<NullWritable> values, Context context
        ) throws IOException, InterruptedException {
            // Output only a single instance of the same key, we don't care about the values
            context.write(key, NullWritable.get());
        }
    }

    public int run(String[] args) throws Exception {
        String inputDir = null;
        String mentionsFile = null;
        String nameVariants = null;
        String outputDir = null;
        boolean useSynsets = false;

        Configuration conf = getConf();


        // parse arguments
        for (int i = 0; i < args.length; ++i) {
            try {
                if ("-i".equals(args[i])) {
                    inputDir = args[++i];
                } else if ("-m".equals(args[i])) {
                    mentionsFile = args[++i];
                } else if ("-nv".equals(args[i])) {
                    nameVariants = args[++i];
                } else if ("-o".equals(args[i])) {
                    outputDir = args[++i];
                } else if ("-synsets".equals(args[i])) {
                    useSynsets = true;
                }
            } catch (ArrayIndexOutOfBoundsException except) {
                System.out.println("ERROR: Required parameter missing from "
                        + args[i - 1]);
                return -1;
            }
        }

        if (inputDir == null || mentionsFile == null || outputDir == null || nameVariants == null) {
            return -1;
        }
        conf.set(CONF_USE_FULLTEXT, Boolean.toString(useSynsets));
        conf.set(CONF_MENTIONS_DIR, mentionsFile);
        conf.set(CONF_VARIANT_DIR, nameVariants);


        Job job = new Job(conf, "KBA - Extract vocabulary" +
                (!useSynsets ? " - raw POS verbs" : " - Wordnet hypernyms");

        job.setJarByClass(VocabularyExtractor.class);

        job.setInputFormatClass(SequenceFileInputFormat.class);
        job.setMapperClass(MyMapper.class);
        job.setMapOutputKeyClass(Text.class);
        job.setMapOutputValueClass(NullWritable.class);
        job.setReducerClass(MyReducer.class);
        job.setOutputFormatClass(TextOutputFormat.class);
        job.setOutputKeyClass(Text.class);
        job.setOutputValueClass(NullWritable.class);
        job.setNumReduceTasks(1);

        FileInputFormat.addInputPath(job, new Path(inputDir));
        FileSystem.get(conf).delete(new Path(outputDir), true);
        TextOutputFormat.setOutputPath(job, new Path(outputDir));

        return job.waitForCompletion(true) ? 0 : 1;
    }


    public static void main(String[] args) throws Exception {
        int res = ToolRunner.run(new Configuration(), new VocabularyExtractor(),
                args);
        System.exit(res);
    }
}
