package no.ntnu.trec.kba.features;

import edu.stanford.nlp.ie.util.RelationTriple;
import edu.stanford.nlp.process.Morphology;
import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.trec.kba.TopicEntity;
import no.ntnu.trec.kba.bin.FeatureExtractorHadoop;
import no.ntnu.utils.HadoopArguments;
import no.ntnu.utils.NlpHelper;
import no.ntnu.utils.WordnetHelper;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.log4j.Logger;
import streamcorpus.Sentence;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

/**
 * Extract document-representation features using VSM
 *
 * Extraction of tense and clause features is also performed here
 * Use the FeatureOutputCombiner if you want to separate these
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */
public class Verbs extends FeatureGenerator {
	private static final Logger log = Logger.getLogger(Verbs.class);


	private boolean useSynsets;

	private Map<String, Integer> featureLookup = null;

	public void buildFeatureLookup(String vocabFile, FileSystem fs) throws IOException {
		Path vocabPath = new Path(vocabFile);
		BufferedReader input = new BufferedReader(new InputStreamReader(fs.open(vocabPath)));

		featureLookup = new HashMap<String, Integer>();

		int curFeatIdx = 1;
		String line;
		while ((line = input.readLine()) != null) {
			featureLookup.put(line, curFeatIdx);
			 curFeatIdx++;
		}
	}

	public Verbs(Configuration config) throws IOException {
		super(config);
		FileSystem fs =	FileSystem.get(config);

        buildFeatureLookup(HadoopArguments.getArgFromConfig(config,"Fv"), fs);
		useSynsets = Boolean.valueOf(HadoopArguments.getArgFromConfig(config,"syngroups"));
		log.info("Using synsets? " + Boolean.toString(useSynsets));
	}

	@Override
	public String getFeatureVector(StreamDocument doc, TopicEntity entity) {
		HashSet<Integer> trueFeatures = new HashSet<Integer>();

		int numTimesObject = 0;
		int numTimesSubject = 0;
		int numPastTense =  0;
		int numPresentTense = 0;

		// Get only the sentences that mention the entity
		List<Sentence> sentences = doc.getSentences();
		sentences = NlpHelper.getMentioningSentences(sentences, entity.getNameVariantsPattern());
		HashSet<String> featureNames = new HashSet<String>();
		for (Sentence sent : sentences) {
			Map<String, String> verbs;
			Pattern myNamePattern = entity.getNameVariantsPattern();
			// Extract only verbs that appear in relation triples
			List<RelationTriple> triples = NlpHelper.getMentioningTriples(sent, myNamePattern);

			HashSet<String> seenRelations = new HashSet<String>();

			// Add to subject/object count features
			for (RelationTriple triple : triples) {
				if (seenRelations.contains(triple.relationGloss()))
					continue; // Skip duplicate

				if (myNamePattern.matcher(triple.objectGloss().toLowerCase()).find()) {
					numTimesObject += 1;
				}
				if (myNamePattern.matcher(triple.subjectGloss().toLowerCase()).find()) {
					numTimesSubject += 1;
				}
				seenRelations.add(triple.relationGloss());
			}

			verbs = NlpHelper.verbsAndTagsFromTriples(triples);

			for (Map.Entry<String, String> entry : verbs.entrySet()) {
				if (!useSynsets) {
					// Feature vector of only lemmatized verbs
					featureNames.add(Morphology.lemmaStatic(entry.getKey(), entry.getValue()));
				} else {
					// Feature vector of hypernyms from Wordnet
					for (String hypernymOffset : WordnetHelper.getVerbHypernyms(entry.getKey(), entry.getValue())) {
						featureNames.add(hypernymOffset);
					}
				}

				// Add to tense features
				if (NlpHelper.verbTagIsPastTense(entry.getValue())) {
					numPastTense++;
				} else {
					// If it is not past tense then the verb has to be present tense (or base form)
					numPresentTense++;
				}
			}
		}

		for (String name : featureNames) {
			Integer featureIdx = featureLookup.get(name);
			if (featureIdx != null) {
				trueFeatures.add(featureIdx);
			}
		}

		// Build the feature vector
		StringBuilder featureVectorBuilder = new StringBuilder();

		// Add verb features
		for (int idx = 0; idx < featureLookup.size(); idx++) {
			if (idx > 0) {
				featureVectorBuilder.append("\t");
			}
			boolean exists = trueFeatures.contains(idx);
			featureVectorBuilder.append(useSynsets ? "h" : "v");
			featureVectorBuilder.append(idx+1);
			featureVectorBuilder.append(":");
			featureVectorBuilder.append(exists ? 1 : 0);
		}

		// Add other features
		featureVectorBuilder.append("\ttensePast:");
		featureVectorBuilder.append(numPastTense);
		featureVectorBuilder.append("\ttensePresent:");
		featureVectorBuilder.append(numPresentTense);
		featureVectorBuilder.append("\tentObj:");
		featureVectorBuilder.append(numTimesObject);
		featureVectorBuilder.append("\tentSub:");
		featureVectorBuilder.append(numTimesSubject);
		featureVectorBuilder.append("\tentTenseSkew:");
		featureVectorBuilder.append(numPresentTense-numPastTense);
		featureVectorBuilder.append("\tentSubSkew:");
		featureVectorBuilder.append(numTimesSubject-numTimesObject);
		
		return featureVectorBuilder.toString();
	}

}
