package no.ntnu.trec.kba.features;

import edu.stanford.nlp.ie.util.RelationTriple;
import edu.stanford.nlp.process.Morphology;
import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.trec.kba.TopicEntity;
import no.ntnu.utils.HadoopArguments;
import no.ntnu.utils.NlpHelper;
import no.ntnu.utils.WordnetHelper;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.log4j.Logger;
import org.deeplearning4j.models.embeddings.loader.WordVectorSerializer;
import org.deeplearning4j.models.embeddings.wordvectors.WordVectors;
import org.nd4j.linalg.api.ndarray.INDArray;
import streamcorpus.Sentence;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.DecimalFormat;
import java.util.*;
import java.util.regex.Pattern;

/**
 * Extract document-representation features using mean of word-vectors representation
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */
public class VerbsEmbedding extends FeatureGenerator {
	private static final Logger log = Logger.getLogger(VerbsEmbedding.class);
	private static DecimalFormat df = new DecimalFormat("0.000000");


	private String zeroResult;


	private WordVectors wordVectors;

	public VerbsEmbedding(Configuration config) throws IOException {
		super(config);

		wordVectors = WordVectorSerializer.loadTxtVectors(
				new File(HadoopArguments.getArgFromConfig(config,"Fv")));

		zeroResult = "";
		for (int i = 0; i < wordVectors.getWordVector("the").length; i++) {
			if (i != 0) {
				zeroResult += "\t";
			}
			zeroResult += "vE" + i + ":0";
		}
	}

	@Override
	public String getFeatureVector(StreamDocument doc, TopicEntity entity) {

		// Get only the sentences that mention the entity
		List<Sentence> sentences = doc.getSentences();
		sentences = NlpHelper.getMentioningSentences(sentences, entity.getNameVariantsPattern());
		HashSet<String> featureNames = new HashSet<String>();
		for (Sentence sent : sentences) {
			Map<String, String> verbs;
			Pattern myNamePattern = entity.getNameVariantsPattern();
			// Extract only verbs that appear in relation triples
			List<RelationTriple> triples = NlpHelper.getMentioningTriples(sent, myNamePattern);

			HashSet<String> seenRelations = new HashSet<String>();

			// Add to subject/object count features
			for (RelationTriple triple : triples) {
				if (seenRelations.contains(triple.relationGloss()))
					continue; // Skip duplicate

				seenRelations.add(triple.relationGloss());
			}

			verbs = NlpHelper.verbsAndTagsFromTriples(triples);

			for (Map.Entry<String, String> entry : verbs.entrySet()) {
				// Feature vector of only lemmatized verbs
				featureNames.add(Morphology.lemmaStatic(entry.getKey(), entry.getValue()));
			}
		}

		INDArray vec;
		try {
			vec = wordVectors.getWordVectorsMean(new ArrayList<String>(featureNames));
		} catch (IllegalArgumentException e) {
			return zeroResult;
		}

		// Build the feature vector
		StringBuilder featureVectorBuilder = new StringBuilder();

		// Add verb features
		for (int idx = 0; idx < vec.columns(); idx++) {
			if (idx > 0) {
				featureVectorBuilder.append("\t");
			}
			featureVectorBuilder.append("vE");
			featureVectorBuilder.append(idx+1);
			featureVectorBuilder.append(":");
			featureVectorBuilder.append(df.format(vec.getDouble(idx)));
		}
		
		return featureVectorBuilder.toString();
	}

}
