package no.ntnu.trec.kba.trainingset;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.SequenceFile.Reader;

public class GenerateArffFileHadoop {
	private HashMap<String, Integer> annotationsTraining;

	private BufferedWriter trainingArffwriter;
	private BufferedWriter testingArffwriter;
	private BufferedWriter idswriter;

	private HashMap<String, String> attributeTypes = null;
	private ArrayList<String> attributes = null;

	private void getAttributeTypes(String attrTypeFile) throws IOException {
		List<String> lines = Files.readAllLines(Paths.get(attrTypeFile));
		attributeTypes = new HashMap<String, String>();

		for (String line : lines) {
			String[] split = line.split("\t");
			attributeTypes.put(split[0], split[1]);
		}
	}

	HashMap<String, Integer> loadAnnotations(String annotFile, Set<Integer> pos, Set<Integer> neg) throws IOException {
		HashMap<String, Integer> list = new HashMap<String, Integer>();
		List<String> lines = Files.readAllLines(Paths.get(annotFile));

		
		for (String line : lines) {
			if (line.startsWith("#"))
				continue;
			String[] split = line.split("\t");
			String mention = split[3];
			String streamid = split[2];
			Integer annotation = Integer.decode(split[5]);
			int classifierClass = 0;
			if (pos == null) {
				list.put(streamid + "\t" + mention, 0);
				continue;
			}
			if (pos.contains(annotation)) {
				classifierClass = 1;
			} else if (!neg.contains(annotation)) {
				continue;
			}
			
			list.put(streamid + "\t" + mention, classifierClass);
		}
		
		return list;
	}

	private void buildAttributeArray(String firstLine) {
		attributes = new ArrayList<String>();

		String[] split = firstLine.split("\t");
		for (String feat : split) {
			String[] nameVal= feat.split(":");
			attributes.add(nameVal[0]);
		}
	}
	
	/**
	 * Generate Arff header with feature names and types (i.e., "@ATTRIBUTE"
	 * lines)
	 * 
	 * @return
	 */
	public String getArffHeader() {
		String output = "@RELATION trec-kba\n";
		for (int i = 0; i < attributes.size(); i++) {
			output += "@ATTRIBUTE " + attributes.get(i) + " NUMERIC\n";
		}
		output += "@ATTRIBUTE id STRING\n";
		output += "@ATTRIBUTE class {0,1}\n\n";
		output  += "@DATA\n";
		return output;
	}
	
	public String getArffVector(String vector) throws Exception {
		HashMap<String, String> lookup = new HashMap<String, String>();
		String[] split = vector.split("\t");
		for (String feat : split) {
			String[] nameVal;
			nameVal = feat.split(":");
			lookup.put(nameVal[0], nameVal[1]);
		}
		String out = "";
		for (String name : attributes) {
			String val = lookup.get(name);
			if (val == null) {
				throw new Exception("Error: Missing feature " + name + ". Feature vector was '" + vector + "'");
			}
			
			if (val.equals("NaN")) {
				val = "?";
			}
			out += val + ",";
		}
		return out;
	}
	
	public void processFeatureVector(String key, String featureVector) throws Exception {
		String out = getArffVector(featureVector);
		testingArffwriter.write(out + "?\n");
		idswriter.write(key + "\n");

		Integer target = annotationsTraining.get(key);
		if (target != null) {
			out = getArffVector(featureVector);
			trainingArffwriter.write(out + "'" + key + "'," + target.toString() + "\n");
		}
	}
	
	public void run(String annotFileTraining, String sequenceFiles, String outputDir,
			Set<Integer> pos, Set<Integer> neg) throws Exception {
		annotationsTraining = loadAnnotations(annotFileTraining, pos, neg);

		trainingArffwriter = new BufferedWriter(new FileWriter(outputDir + "/training.arff"));
		testingArffwriter = new BufferedWriter(new FileWriter(outputDir + "/testing.arff"));
		idswriter = new BufferedWriter(new FileWriter(outputDir + "/testing.ids"));

		Configuration conf = new Configuration();
		FileSystem fs = FileSystem.get(conf);
		
		for (FileStatus file : fs.listStatus(new Path(sequenceFiles))) {
			if (!file.getPath().getName().startsWith("part")) {
				continue;
			}
			SequenceFile.Reader reader = new SequenceFile.Reader(conf, Reader.file(file.getPath()));
			
			Text key = new Text();
			Text featureVector = new Text();


			while (reader.next(key, featureVector)) {
				if (attributes == null) {
					buildAttributeArray(featureVector.toString());
					String header = getArffHeader();
					testingArffwriter.write(header);
					trainingArffwriter.write(header);
				}
				processFeatureVector(key.toString(), featureVector.toString());
			}
			reader.close();

		}
		testingArffwriter.close();
		trainingArffwriter.close();
		idswriter.close();

	}
	
	static Set<Integer> parseIntList(String str) {
		Set<Integer> s = new HashSet<Integer>();
		String[] tmp = str.split(",");
		for (int i = 0; i < tmp.length; i++) {
			s.add(new Integer(tmp[i]));
		}
		return s;
	}
	
	public static void main(String[] args) throws Exception {
		String attrTypeFile = null;
		String sequenceFiles = null;
		String trainingAnnotFile = null;
		String outputDir = null;
		Set<Integer> positiveClass = new HashSet<Integer>();
		Set<Integer> negativeClass = new HashSet<Integer>();

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-attrTypes".equals(args[i])) {
					attrTypeFile = args[++i];
				} else if ("-atrain".equals(args[i])) {
					trainingAnnotFile = args[++i];
				} else if ("-i".equals(args[i])) {
					sequenceFiles = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				} else if ("-pc".equals(args[i])) {
					positiveClass = parseIntList(args[++i]);
				} else if ("-nc".equals(args[i])) {
					negativeClass = parseIntList(args[++i]);
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
			}
		}

		if (trainingAnnotFile == null || outputDir == null) {
			System.out.println("Missing input, aborting");
			return;
		}

		
		GenerateArffFileHadoop a = new GenerateArffFileHadoop();
		
		a.run(trainingAnnotFile, sequenceFiles, outputDir, positiveClass, negativeClass);
	}
}
