package no.ntnu.trec.kba.trainingset;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.regex.Pattern;

import com.amazonaws.services.dynamodbv2.xspec.S;
import no.ntnu.trec.kba.TopicEntity;
import no.ntnu.utils.NlpHelper;
import no.ntnu.utils.SinglePassClusters;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Text;

import no.ntnu.trec.kba.StreamDocument;
import no.ntnu.utils.EntityNameResolver;
import streamcorpus.Sentence;

/**
 * Generates entity profiles from training data
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */
public class TrainingProfileGenerator extends TrainingSetProcessor {
	private HashMap<String, String> profiles;
	private HashMap<String, TopicEntity> entityProfiles;
	private HashMap<String, SinglePassClusters> entityClusters;
	private String entityNameVarFile;
	
	public TrainingProfileGenerator(String entitiesFile, String annotationsFile, String outputDir, String entityNameVarFile) throws IOException {
		super(entitiesFile, annotationsFile, outputDir);
		this.entityNameVarFile = entityNameVarFile;

		entityProfiles = new HashMap<String, TopicEntity>();
        entityClusters = new HashMap<String, SinglePassClusters>();
		profiles = new HashMap<String, String>();
	}

	private Pattern getNameVariants(String entity) {
		entity = EntityNameResolver.removeNamePrefix(entity);
		if (!entityProfiles.containsKey(entity)) {
			TopicEntity profile = new TopicEntity(entity);
			try {
				profile.addNameVariantsFromFile(
						new Path(entityNameVarFile + "/" + entity),
						FileSystem.get(new Configuration()));
			} catch (Exception e) {
				System.out.println("Could not get name variants from file: " + e.getCause());
			}

			entityProfiles.put(entity, profile);
		}

		return entityProfiles.get(entity).getNameVariantsPattern();
	}

	private SinglePassClusters getEntityClusters(String entity) {
	    if (!entityClusters.containsKey(entity)) {
	        entityClusters.put(entity, new SinglePassClusters(0.9f));
        }

        return entityClusters.get(entity);
    }

	@Override
	void map(Text key, StreamDocument item) {
		if (mentions.containsKey(key.toString())) {
			for (String mention : mentions.get(key.toString())) {
				if (entities.contains(mention) && annotations.get(key + "\t" + mention) > 0) {

					List<Sentence> sentences = NlpHelper.getMentioningSentences(
							item.getSentences(),
							getNameVariants(mention));

					SinglePassClusters entityClusters = getEntityClusters(mention);

					String tmp = "";
					for (Sentence sent : sentences) {
					    SinglePassClusters.Cluster cluster = entityClusters.
                                addDocumentToCluster(
                                        NlpHelper.getFreqMapFromSentence(sent),
                                        item.get(StreamDocument.FIELD_DIR_NAME));

					    // Only add to profile if the document belongs to a new cluster (ie. not a duplicate)
                        if (cluster.getMemberCount() == 1) {
                            tmp += NlpHelper.sentenceToString(sent) + " ";
                        }
					}

					if (!tmp.equals("")) {
						if (profiles.containsKey(mention)) {
							profiles.put(mention, profiles.get(mention).concat(" " + tmp));
						} else {
							profiles.put(mention, tmp);
						}
					}

				}
			}
		}
		
	}

	@Override
	void after() throws IOException {
		for (String entity : entities) {
			String out_entity = EntityNameResolver.removeNamePrefix(entity);
			Files.write(Paths.get(outputDir + "/" + out_entity), profiles.getOrDefault(entity, "").getBytes(), StandardOpenOption.CREATE);
		}
	}
	
	public static void main(String[] args) {
		try {
			TrainingProfileGenerator generator = new TrainingProfileGenerator(args[0], args[1], args[2], args[3]);
			generator.run(args[4]);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

}
