package no.ntnu.utils;

import edu.stanford.nlp.ie.util.RelationTriple;
import edu.stanford.nlp.ling.CoreAnnotations;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.naturalli.NaturalLogicAnnotations;
import edu.stanford.nlp.parser.nndep.DependencyParser;
import edu.stanford.nlp.pipeline.Annotation;
import edu.stanford.nlp.pipeline.StanfordCoreNLP;
import edu.stanford.nlp.process.Morphology;
import edu.stanford.nlp.trees.GrammaticalStructure;
import edu.stanford.nlp.util.CoreMap;
import org.apache.log4j.Logger;
import org.apache.lucene.analysis.core.StopAnalyzer;
import org.deeplearning4j.text.stopwords.StopWords;
import org.nd4j.linalg.api.ndarray.INDArray;
import streamcorpus.Relation;
import streamcorpus.Sentence;
import streamcorpus.Token;

import java.lang.reflect.Array;
import java.util.*;
import java.util.regex.Pattern;

public class NlpHelper {
    private static StanfordCoreNLP _pipeline = null;
    private static Set<String> stopwordLookup = null;
    private static final Logger log = Logger.getLogger(NlpHelper.class);
    private static String[] bannedTokens = {":", "-", "n't", "'s", "``", ",", ",,", ".", "''", "?", "br", "/", "em", "-rrb-", "-lrb-", "|", "span", "em", "div", "li", "p"};
    private static Set<String> sentenceTokenFilter = new HashSet<String>(Arrays.asList(bannedTokens));

    private static StanfordCoreNLP getStanfordPipeline() {
        if (_pipeline == null) {
            Properties props = new Properties();
            props.setProperty("annotators", "tokenize,ssplit,pos,lemma,depparse,natlog,openie");
            _pipeline = new StanfordCoreNLP(props);
        }

        return _pipeline;
    }

    private static Set<String> getStopwordLookup() {
        if (stopwordLookup == null) {
            stopwordLookup = new HashSet<String>(StopWords.getStopWords());
        }

        return stopwordLookup;
    }

    public static String sentenceToString(Sentence sentence) {
        return sentenceToString(sentence, false);
    }

    public static String sentenceToString(Sentence sentence, boolean removeStopwords) {
        List<String> tokenList = sentenceToList(sentence, removeStopwords);
        String buf = "";
        for (String tok : tokenList) {
            buf += tok + " ";
        }
        return buf;
    }

    public static List<String> sentenceToList(Sentence sentence) {
        return sentenceToList(sentence, false);
    }

    public static Set<String> getTokenFilter() {
        return sentenceTokenFilter;
    }

    public static List<String> sentenceToList(Sentence sentence, boolean removeStopwords) {
        List<String> tokenList = new ArrayList<String>();
        for (Token tok : sentence.getTokens()) {
            String tok_text = tok.token.toLowerCase();
            if (!removeStopwords || !getStopwordLookup().contains(tok_text)) {
                tokenList.add(tok.token.toLowerCase());
            }

        }
        return tokenList;
    }

    public static Set<String> getVerbs(Sentence sentence) {
        Set<String> verbs = new HashSet<String>();

        for (Token tok : sentence.getTokens()) {
            if (tokenIsVerb(tok.pos)) {
                verbs.add(tok.token);
            }
        }

        return verbs;
    }

    public static Map<String, String> verbsAndTagsFromSentence(Sentence sentence) {
        Map<String, String> verbs = new HashMap<String, String>();

        for (Token tok : sentence.getTokens()) {
            if (tokenIsVerb(tok.pos)) {
                verbs.put(tok.token, tok.pos);
            }
        }

        return verbs;
    }

    public static List<TaggedWord> serifSentenceToCoreNlp(Sentence sent) {
        List<TaggedWord> coreNlpSentence = new ArrayList<TaggedWord>();
        for (Token tok : sent.getTokens()) {
            coreNlpSentence.add(new TaggedWord(tok.token, tok.pos));
        }

        return coreNlpSentence;
    }

    public static Map<String, String> verbsAndTagsFromTriples(List<RelationTriple> triples) {
        Map<String, String> verbs = new HashMap<String, String>();

        for (RelationTriple triple : triples) {
            for (CoreLabel word : triple.relation) {
                if (verbs.containsKey(word.word())) {
                    continue; // Skip words we have already extracted
                }
                if (tokenIsVerb(word.tag())) {
                    verbs.put(word.word(), word.tag());
                }
            }
        }

        return verbs;
    }

    public static List<RelationTriple> getMentioningTriples(Sentence sent, Pattern namePattern) {
        ArrayList<RelationTriple> mentioningTriples = new ArrayList<RelationTriple>();

        // Very long sentences eat all the memory of Hadoop
        if (sent.getTokens().size() > 50) {
            ArrayList<Token> newList = new ArrayList<Token>();
            for (int i = 0; i < 50; i++) {
                newList.add(sent.getTokens().get(i));
            }
            sent.setTokens(newList);
        }

        StanfordCoreNLP pipeline = getStanfordPipeline();
        String sentStr = NlpHelper.sentenceToString(sent);



        // Remove characters that sometimes causes hang in parser
        sentStr = sentStr.replaceAll("\\|", "");
        sentStr = sentStr.replaceAll("\\$", "");

        Annotation annot = pipeline.process(sentStr);
        for (CoreMap sentence : annot.get(CoreAnnotations.SentencesAnnotation.class)) {
            Collection<RelationTriple> triples =
                    sentence.get(NaturalLogicAnnotations.RelationTriplesAnnotation.class);

            for (RelationTriple triple : triples) {
                String objectSubject = triple.subjectGloss() + " " + triple.objectGloss();
                if (namePattern.matcher(objectSubject.toLowerCase()).find()) {
                    // Relation triple contains mention of entity
                    mentioningTriples.add(triple);
                }
            }
        }


        return mentioningTriples;
    }

    public static boolean tokenIsVerb(String posTag) {
        return posTag.startsWith("VB");
    }

    public static List<Sentence> getMentioningSentences(List<Sentence> sentences, Pattern namePattern) {
        ArrayList<Sentence> matches = new ArrayList<Sentence>();
        for (Sentence sent : sentences) {
            String strSent = sentenceToString(sent);
            if (MentionTextHelper.hasMoreThanOneMention(strSent, namePattern)) {
                matches.add(sent);
            }
        }

        return matches;
    }

    public static Set<String> getTokensetFromSentence(Sentence sentence) {
        Set<String> set = new HashSet<String>();

        for (Token tok : sentence.getTokens()) {
            if (!StopAnalyzer.ENGLISH_STOP_WORDS_SET.contains(tok.token.toLowerCase())) {
                set.add(tok.token.toLowerCase());
            }
        }

        return set;
    }

    public static Map<String, Integer> getFreqMapFromSentences(List<Sentence> sentences) {
        HashMap<String, Integer> map = new HashMap<String, Integer>();

        for (Sentence sent : sentences) {
            for (Token tok : sent.getTokens()) {
                String strTok = tok.token.toLowerCase();
                Integer cnt = map.get(strTok);
                if (cnt == null) {
                    map.put(strTok, 1);
                } else {
                    map.put(strTok, cnt+1);
                }

            }
        }


        return map;
    }

    public static Map<String, Integer> getFreqMapFromSentence(Sentence sentence) {
        HashMap<String, Integer> map = new HashMap<String, Integer>();

        for (Token tok : sentence.getTokens()) {
            String strTok = tok.token.toLowerCase();
            Integer cnt = map.get(strTok);
            if (cnt == 0) {
                map.put(strTok, 1);
            } else {
                map.put(strTok, cnt+1);
            }

        }

        return map;
    }

    public static Set<String> getTokensetFromSentences(List<Sentence> sentences) {
        Set<String> set = new HashSet<String>();

        for (Sentence sent : sentences) {
            for (Token tok : sent.getTokens()) {
                if (!StopAnalyzer.ENGLISH_STOP_WORDS_SET.contains(tok.token.toLowerCase())) {
                    set.add(tok.token.toLowerCase());
                }
            }
        }

        return set;
    }

    public static boolean verbTagIsPastTense(String tag) {
        return tag.equals("VBD") || tag.equals("VBN");
    }

    public static void string2Bow(String str) {

    }
}
