package no.ntnu.utils;

import java.util.*;

/**
 * Implemention of the Single Pass clustering algorithm
 */
public class SinglePassClusters {

    private float similarityThreshold;
    private List<Cluster> clusters;

    public class Cluster {
        List<Set<String>> documents;
        int documentCount;
        String firstDocumentTimestamp;

        Cluster(Set<String> initalDocument, String firstTimestamp) {
            documents = new ArrayList<Set<String>>();
            documentCount = 0;
            addDocument(initalDocument, true);
            firstDocumentTimestamp = firstTimestamp;
        }

        /**
         * Add a document to this cluster
         *
         * If the document distance is below the contribution threshold it is simply ignored
         * @param document Document to add
         */
        public void addDocument(Set<String> document, boolean addToDefiniton) {
            documentCount++;
            if (addToDefiniton)
                documents.add(document);
        }

        /**
         * Add a document to this cluster
         *
         * If the document distance is below the contribution threshold it is simply ignored
         */
        public int getMemberCount() {
            return documentCount;
        }

        /**
         * Distance measure between two documents (Dice coefficient)
         * @param doc1 First document
         * @param doc2 Second document
         * @return Similarity between documents using Dice coefficient
         */
        public float doc2docSimilarity(Set<String> doc1, Set<String> doc2) {
            int intersectionSize = 0;

            for (String term : doc1) {
                if (doc2.contains(term)) {
                    intersectionSize++;
                }
            }

            return (2.f*intersectionSize)/(doc1.size()+doc2.size());
        }

        /**
         * Distance measure between two documents
         * @param doc Document to compare to cluster
         * @return Similarity between documents using Dice coefficient
         */
        public float doc2clusterSimilarity(Set<String> doc) {
            float sim = 0;

            for (Set<String> clusterDoc : documents) {
                sim += doc2docSimilarity(doc, clusterDoc);
            }

            return sim/documents.size();
        }


        /**
         * Get the timestamp of the first document added to the cluster
         * @return Timestamp of first document
         */
        public String getTimestamp() {
            return firstDocumentTimestamp;
        }
    }

    /**
     * Initialize new cluster container
     * @param similarityThreshold Similarity threshold for including a document in a cluster
     */
    public SinglePassClusters(float similarityThreshold) {
        this.similarityThreshold = similarityThreshold;
        clusters = new ArrayList<Cluster>();
    }

    /**
     * Add the document to the closest cluster and return the cluster it was added to
     *
     * If no existing cluster was found a new one is created
     *
     * @param document Document Document to add to a cluster
     * @param timestamp Timestamp of document
     * @return The cluster the document was added to
     */
    public Cluster addDocumentToCluster(Set<String> document, String timestamp) {
        Cluster myClosestCluster = null;
        float myClosestSim = Float.MIN_VALUE;

        // Find closest cluster
        for (Cluster cluster : clusters) {
            float candidateSim = cluster.doc2clusterSimilarity(document);

            if (candidateSim > myClosestSim) {
                myClosestSim = candidateSim;
                myClosestCluster = cluster;
            }
        }

        // Create new cluster if no appropriate existing cluster exists
        if (myClosestCluster == null || myClosestSim < similarityThreshold) {
            Cluster newCluster = new Cluster(document, timestamp);
            clusters.add(newCluster);

            return newCluster;
        }

        // Add to existing cluster
        myClosestCluster.addDocument(document, false);

        return myClosestCluster;
    }
}
