package no.ntnu.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.builder.CompareToBuilder;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.log4j.Logger;
import org.apache.lucene.index.AtomicReader;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.FieldInfo;
import org.apache.lucene.index.FieldInfos;
import org.apache.lucene.index.Fields;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.Terms;
import org.apache.lucene.index.TermsEnum;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.util.BytesRef;

/**
 * Collects term frequency statistics for fast access.
 * 
 * For now, statistics are computed for a single index field.
 * 
 * @author Heri Ramampiaro <heri@ntnu.no>
 * @author Krisztian Balog <krisztian.balog@idi.ntnu.no>
 */

public class TermFreq {
	private static Logger logger = Logger.getLogger(TermFreq.class);

	private String fieldName;

	private Map<String, Freq> termMap;

	/** total number of documents */
	private int docNum;

	/** sum of term cf values (for normalization purposes) */
	private long sumCf;

	public TermFreq(String termFile, FileSystem fs, String fieldName) throws IOException {
		this.fieldName = fieldName;
		
		try {
			buildTermMap(termFile, fs);
		} catch (Exception e) {
			throw new IOException("Failed to parse term file: " + e.getMessage());
		}
	}

	/**
	 * Building the term map from the index.
	 * @throws IOException 
	 * @throws NumberFormatException 
	 */
	public void buildTermMap(String termFile, FileSystem fs) throws NumberFormatException, IOException {

		termMap = Collections.synchronizedMap(new HashMap<String, Freq>());
		sumCf = 0;
		
		Path termPath = new Path(termFile);
		
		BufferedReader input = new BufferedReader(new InputStreamReader(fs.open(termPath)));

		Path tFile = new Path(termFile);
		if (!fs.exists(tFile)) {
			throw new IOException("Could not open term file");
		}
		
		docNum = Integer.parseInt(input.readLine());
		
		String line;
		while ((line = input.readLine()) != null) {
			if (line.length() == 0)
				continue;
			String[] split = line.split("\t");
			// Document frequency
			int df = Integer.parseInt(split[1]);
			// Collection frequency
			long cf = Long.parseLong(split[2]);
			sumCf += cf;

			// we consider a single field only, so we must have not seen
			// this term before
			Freq freq = new Freq(split[0], cf, df);
			termMap.put(split[0], freq);
		}

	}

	/**
	 * Get IDF of a given term
	 * 
	 * @param term
	 * @return
	 */
	public double getIDF(String term) {
		int df = 0;
		if (termMap.containsKey(term)) {
			df = termMap.get(term).df;
			return Math.log((double) docNum / df);
		}
		// Make the assumtion that a non-found df is somewhere between 1-10
		return Math.log((double) docNum / 5);
	}

	/**
	 * Get normalized CF of a given term
	 * 
	 * @param term
	 * @return
	 */
	public double getCFNormalized(String term) {
		long cf = 0;
		if (termMap.containsKey(term)) {
			cf = termMap.get(term).cf;
		} else {
			// Assume that a non-found cf is between 1-10, estimate 5
			cf = 5;
		}
		return cf / sumCf;
	}

	/**
	 * Get the freq-statistics for a given term
	 * 
	 * @param term
	 *            the term
	 * @return Freq
	 */
	public Freq getTermFreq(String term) {
		if (termMap.containsKey(term)) {
			return termMap.get(term);
		} else {
			return new Freq(term, 5, 5);
		}
	}

	/**
	 * Get average document length
	 * 
	 * @return
	 */
	public double getAvgDl() {
		return (double) sumCf / docNum;
	}
	
	public long getTotalCf() {
		return sumCf;
	}

	public static void main(String[] pArgs) throws Exception {
		logger.info("Testing this class.");

		/*String indexDir = "/Users/kbalog/work/trec-kba/index-dbpedia+training";
		TermFreq tf = new TermFreq(indexDir, "body_cleansed");

		String text = "Aharon Barak (born 16 September 1936) is a Professor's of Law at the Interdisciplinary\n"
				+ "Center in Herzliya and a lecturer in law at the Hebrew University of Jerusalem, the Yale Law School, Georgetown\n\n"
				+ " University Law Center, and the University of Toronto Faculty of Law."
				+ "Barak was President of the Supreme Court of Israel from 1995 to 2006. Prior to that, he served"
				+ " as a Justice on the Supreme Court of Israel (1978-95), as the Attorney General of Israel (1975-78)"
				+ ", and as the Dean of the Law Faculty of the Hebrew University of Jerusalem (1974-75)";
		String at[] = TextAnalysis.analyze(text);

		System.out.println("Frequency information: [df, cf]:");
		for (String t : at) {
			Freq f = tf.getTermFreq(t);
			System.out.println("Frequency of " + f.term + ": [" + f.df + ", "
					+ f.cf + "]");
		}*/
	}

	/**
	 * An internal class holding frequency values for a term
	 * 
	 * @author Heri Ramampiaro <heri@ntnu.no>
	 * 
	 */
	public static class Freq implements Comparable<Object> {
		public String term;
		public int df;
		public long cf;

		public Freq(String term, long cf, int df) {
			this.term = term;
			this.cf = cf;
			this.df = df;

		}

		public double idf(long docNum) {
			return Math.log(docNum / (double) df);
		}

		/**
		 * 
		 * Can be useful if we like to do term sorting
		 */
		public int compareTo(Object o) {
			if (o instanceof Freq) {
				Freq oFreq = (Freq) o;
				return new CompareToBuilder().append(cf, oFreq.cf)
						.append(term, oFreq.term).toComparison();
			} else {
				return 0;
			}
		}
	}
}