package no.ntnu.utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Methods for computing similarity between text
 * 
 * @author Heri Ramampiaro <heri@ntnu.no>
 * @author Krisztian Balog <krisztian.balog@idi.ntnu.no>
 * 
 */
public class TextSimilarity {


	/**
	 * Calculate the Jaccard similarity between two documents
	 * 
	 * @param text1
	 * @param text2
	 * @return
	 */
	public static double jaccardSimilarity(String[] text1, String[] text2) {
		
		// Creating the union set
		Set<String> union_set = new HashSet<String>(Arrays.asList(text1));
		union_set.addAll(Arrays.asList(text2));

		// Creating the intersection set
		Set<String> intersection_set = new HashSet<String>(Arrays.asList(text1));
		intersection_set.retainAll(Arrays.asList(text2));
		
		// return JaccardSimilarity
		return (double) (intersection_set.size()) / (double) (union_set.size());

	}

	/**
	 * Get term frequency map
	 * 
	 * @param text
	 * @return
	 */
	private static Map<String, Integer> getTermFreqMap(String[] text) {
		Map<String, Integer> freq = new HashMap<String, Integer>();
		for (String term : text) {
			int f = 1;
			if (freq.containsKey(term))
				f += freq.get(term);
			freq.put(term, f);
		}
		return freq;
	}
	
	/**
	 * Get TF-IDF term vector
	 * 
	 * @param text
	 * @param termFreq
	 * @return
	 */
	private static Map<String, Double> getTFIDFVector(String[] text,
			TermFreq termFreq) {
		
		Map<String, Integer> freq = getTermFreqMap(text);
		int freqSum = 0;
		for (int tf : freq.values()) {
			freqSum += tf;
		}
		Map<String, Double> tfidf = new HashMap<String, Double>();

		for (String term : freq.keySet()) {
			double tf = (double) freq.get(term) / freqSum;
			tfidf.put(term, tf * termFreq.getIDF(term));
		}

		return tfidf;
	}

	/**
	 * Compute the cosine similarity between two documents
	 * 
	 * @param text1
	 * @param text2
	 * @param termFreq
	 * @return
	 */
	public static double cosineSimilarity(String[] text1, String[] text2,
			TermFreq termFreq) {

		// Get TF-IDF vectors
		Map<String, Double> tfidf1 = getTFIDFVector(text1, termFreq);
		Map<String, Double> tfidf2 = getTFIDFVector(text2, termFreq);

		// Build the vector
		Set<String> allTerms = new HashSet<String>();
		allTerms.addAll(tfidf1.keySet());
		allTerms.addAll(tfidf2.keySet());

		double scalar = 0, norm1 = 0, norm2 = 0;

		// Compute the dot product between the two vectors
		for (String t : allTerms) {
			if (tfidf1.containsKey(t) && tfidf2.containsKey(t))
				scalar += tfidf1.get(t) * tfidf2.get(t);
		}

		// Compute the norms of both of the vectors
		for (String t : tfidf1.keySet())
			norm1 += Math.pow(tfidf1.get(t), 2);
		for (String t : tfidf2.keySet())
			norm2 += Math.pow(tfidf2.get(t), 2);

		// Return the cosine similarity
		return scalar / Math.sqrt(norm1 * norm2);
	}

	/**
	 * Get dirichlet-smoothed Language Model for a document. The smoothing param
	 * is set to the average document length in the collection.
	 * 
	 * @param text
	 * @param termFreq
	 * @return
	 */
	private static Map<String, Double> getDirichletSmoothedLM(String[] text,
			TermFreq termFreq) {
		
		Map<String, Integer> freq = getTermFreqMap(text);
		Map<String, Double> lm = new HashMap<String, Double>();
		// smoothing param is set to the avg dl
		double mu = termFreq.getAvgDl(); 

		for (String term : freq.keySet()) {
			// P(t|\theta_d) = \frac{n(t,d) + \mu * P(t))}{|d| + \mu}
			double pt = termFreq.getCFNormalized(term);
			double ptd = ((double) freq.get(term) + mu * pt) / ((double) text.length + mu);
			lm.put(term, ptd);
		}

		return lm;
	}

	/**
	 * Computes the KL-divergence between two documents
	 * 
	 * @param text1
	 * @param text2
	 * @param termFreq
	 * @return
	 */
	public static double klDivergence(String[] text1, String[] text2,
			TermFreq termFreq) {

		// Get LMs
		Map<String, Double> lm1 = getDirichletSmoothedLM(text1, termFreq);
		Map<String, Double> lm2 = getDirichletSmoothedLM(text2, termFreq);

		double kl = 0.0;

		for (String t : lm1.keySet()) {
			// Compute the prob. for the token
			double p1 = (double) lm1.get(t);

			if (lm2.containsKey(t)) {
				double p2 = (double) lm2.get(t); 
				kl += p1 * (Math.log((p1) / (p2)));
			}
		}
		return kl;
	}

	public static void main(String[] args) {

		/*String indexDir = "/Users/kbalog/work/trec-kba/index-dbpedia+training";
		TermFreq termFreq = new TermFreq(indexDir, "body_cleansed");

		String text1 = "Aharon Barak (born 16 September 1936) is a Professor of Law at the Interdisciplinary"
				+ "Center in Herzliya and a lecturer in law at the Hebrew University of Jerusalem, the Yale Law School, Georgetown"
				+ " University Law Center, and the University of Toronto Faculty of Law."
				+ "Barak was President of the Supreme Court of Israel from 1995 to 2006. Prior to that, he served"
				+ " as a Justice on the Supreme Court of Israel (1978�95), as the Attorney General of Israel (1975�78)"
				+ ", and as the Dean of the Law Faculty of the Hebrew University of Jerusalem (1974�75)";

		String text2 = "The Honorable Aharon Barak is a Visiting Professor of Law and the Gruber Global Constitutionalism"
				+ " Fellow at Yale Law School. Originally appointed Justice of the Supreme Court of Israel in 1978, Justice "
				+ "Barak served as Supreme Court President from 1995 until 2006. He studied law at the Hebrew University of "
				+ "Jerusalem where, after completing a term of service in the Israeli Defense Forces, he received his Doctorate "
				+ "in Law in 1963.";

		double jacc = jaccardSimilarity(TextAnalysis.analyze(text1),
				TextAnalysis.analyze(text2));
		double cos = cosineSimilarity(TextAnalysis.analyze(text1),
				TextAnalysis.analyze(text2), termFreq);
		double kl = klDivergence(TextAnalysis.analyze(text1),
				TextAnalysis.analyze(text2), termFreq);

		System.out.println("Similarities: \n\tJaccard=" + jacc + "\n\tCosine="
				+ cos + "\n\tKL-divergence=" + kl);*/
	}

}
