package no.ntnu.utils.classification;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.SequenceFile.Reader;
import org.apache.hadoop.io.Text;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * Convert feature output file from feature extractors into libSVM format for use with ranking frameworks
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 */
public class FeatureOutput2LibSVM {
    private HashMap<String, Integer> annotationsTraining;

    private HashMap<String, ArrayList<String>> entityTrainingVectors = new HashMap<String, ArrayList<String>>();
    private HashMap<String, ArrayList<String>> entityTestingVectors = new HashMap<String, ArrayList<String>>();
    private HashMap<String, ArrayList<String>> entityTestingIds = new HashMap<String, ArrayList<String>>();
    private ArrayList<String> entityVector = null;
    private HashMap<String, Integer> entityIds = new HashMap<String, Integer>();
    private int idCounter = 1;

    HashMap<String, Integer> loadAnnotations(String annotFile) throws IOException {
        HashMap<String, Integer> list = new HashMap<String, Integer>();
        List<String> lines = Files.readAllLines(Paths.get(annotFile));

        for (String line : lines) {
            if (line.startsWith("#"))
                continue;
            String[] split = line.split("\t");
            String mention = split[3];
            String streamid = split[2];
            Integer annotation = Integer.decode(split[5]);

            list.put(streamid + "\t" + mention, annotation);
        }

        return list;
    }


    public HashMap<String, String> getFeatures(String vector) throws Exception {
        HashMap<String, String> lookup = new HashMap<String, String>();
        String[] split = vector.split("\t");
        for (String feat : split) {
            String[] nameVal;
            nameVal = feat.split(":");
            if (nameVal[1].equals("?") || nameVal[1].equals("NaN"))
                nameVal[1] = "0"; // RankLib does not like unknown features
            lookup.put(nameVal[0], nameVal[1]);
        }

        return lookup;
    }

    public void processFeatureVector(String key, String featureVector) throws Exception {
        String[] split = key.split("\t");
        String entity = split[1];
        HashMap<String, String> featureLookup = getFeatures(featureVector);

        Integer myId = entityIds.get(entity);
        if (myId == null) {
            myId = idCounter;
            entityIds.put(entity, myId);
            idCounter++;
        }

        // New entity, we need to initialize its vectors
        if (entityVector == null) {
            entityVector = new ArrayList<String>();
            // Feature vector of the first instance becomes defacto feature vector for the whole run
            for (String feat : featureLookup.keySet()) {
                entityVector.add(feat);
            }

            String logString = "Using " + entityVector.size() + " features.";
            System.out.println(logString);
        }

        if (entityTrainingVectors.get(entity) == null) {
            System.out.println("New entity: " + entity);
            entityTrainingVectors.put(entity, new ArrayList<String>());
            entityTestingIds.put(entity, new ArrayList<String>());
            entityTestingVectors.put(entity, new ArrayList<String>());
        }

        // Get entity's training and testing vectors
        ArrayList<String> entityTestingVector = entityTestingVectors.get(entity);
        ArrayList<String> entityIds = entityTestingIds.get(entity);


        // Build the SVMRank feature vector for this instance
        String svmRankInst = "";
        for (int i  = 0; i < entityVector.size(); i++) {
            String val = featureLookup.get(entityVector.get(i));

            if (val == null) {
                System.out.println("Warning: Missing feature " + entityVector.get(i) + " in " + key);
                val = "0";
            }

            svmRankInst += " " + (i+1) + ":" + val;
        }

        if (entityVector.size() != featureLookup.size()) {
            System.out.println("Warning: Entity feature count not equal: " + entityVector.size() + " vs " + featureLookup.size());
        }
        entityTestingVector.add("0" + svmRankInst);
        entityIds.add(key);

        Integer training_rank = annotationsTraining.get(key);
        if (training_rank != null) {
            ArrayList<String> entityTrainingVector = entityTrainingVectors.get(entity);
            entityTrainingVector.add(Integer.toString(training_rank+1) + svmRankInst);
        }
    }


    public void run(String annotFileTraining, String sequenceFiles, String outputDir) throws Exception {
        annotationsTraining = loadAnnotations(annotFileTraining);

        Configuration conf = new Configuration();
        FileSystem fs = FileSystem.get(conf);

        // Build lists of SVMRank instance lines for each entity
        for (FileStatus file : fs.listStatus(new Path(sequenceFiles))) {
            if (!file.getPath().getName().startsWith("part")) {
                continue;
            }
            Reader reader = new Reader(conf, Reader.file(file.getPath()));

            Text key = new Text();
            Text featureVector = new Text();

            while (reader.next(key, featureVector)) {
                processFeatureVector(key.toString(), featureVector.toString());
            }
            reader.close();
        }

        // Output SVMRank instance lines for each entity separately

        BufferedWriter trainingWriter = new BufferedWriter(new FileWriter(outputDir + "/data.train"));
        BufferedWriter trainingGroupsWriter = new BufferedWriter(new FileWriter(outputDir + "/data.train.groups"));
        BufferedWriter testingWriter = new BufferedWriter(new FileWriter(outputDir + "/data.test"));
        BufferedWriter testingGroupsWriter = new BufferedWriter(new FileWriter(outputDir + "/data.test.groups"));
        BufferedWriter testingIdWriter = new BufferedWriter(new FileWriter(outputDir + "/data.test.ids"));

        for (String entity : entityTrainingVectors.keySet()) {
            System.out.println("Processing " + entity + "...");


            ArrayList<String> entityTrainingVector = entityTrainingVectors.get(entity);
            ArrayList<String> entityTestingVector = entityTestingVectors.get(entity);
            ArrayList<String> testingIds = entityTestingIds.get(entity);

            testingGroupsWriter.write(entityTestingVector.size() + "\n" );
            trainingGroupsWriter.write(entityTrainingVector.size() + "\n");
            for (String inst : entityTestingVector) {
                testingWriter.write(inst + "\n");

            }
            for (String inst : testingIds) {
                testingIdWriter.write(inst + "\n");

            }
            for (String inst : entityTrainingVector) {
                trainingWriter.write(inst + "\n");
            }
        }

        // Print feature vector
        System.out.println("Feature vector list in order:");
        for (String feat : entityVector) {
            System.out.println(feat);
        }

        trainingWriter.close();
        testingWriter.close();
        testingGroupsWriter.close();
        trainingGroupsWriter.close();
        testingIdWriter.close();

    }

    public static void main(String[] args) throws Exception {
        String sequenceFiles = null;
        String trainingAnnotFile = null;
        String outputDir = null;

        // parse arguments
        for (int i = 0; i < args.length; ++i) {
            try {
                if ("-a".equals(args[i])) {
                    trainingAnnotFile = args[++i];
                } else if ("-i".equals(args[i])) {
                    sequenceFiles = args[++i];
                } else if ("-o".equals(args[i])) {
                    outputDir = args[++i];
                }
            } catch (ArrayIndexOutOfBoundsException except) {
                System.out.println("ERROR: Required parameter missing from "
                        + args[i - 1]);
            }
        }

        if (trainingAnnotFile == null || outputDir == null || sequenceFiles == null) {
            System.out.println("Missing input, aborting");
            return;
        }


        FeatureOutput2LibSVM a = new FeatureOutput2LibSVM();

        a.run(trainingAnnotFile, sequenceFiles, outputDir);
    }
}
