package no.ntnu.utils.classification;

import no.ntnu.utils.EntityNameResolver;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.SequenceFile.Reader;
import org.apache.hadoop.io.Text;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.*;

/**
 * Use Ranklib to train entity-dependent ranking model and produce runfile
 *
 * Ranklib must be located in lib/RankLib-2.9.jar
 *
 * Parameters:
 * a: Path to file containing training data
 * i: Path to file containing entity-document feature vectors
 * n: Run name
 * s: Ranklib settings (Default Random Forests with default hyperparams)
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 *
 */

public class RankLibRanker {
	private HashMap<String, Integer> annotationsTraining;

	private HashMap<String, ArrayList<String>> entityTrainingVectors = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<String>> entityTestingVectors = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<String>> entityTestingIds = new HashMap<String, ArrayList<String>>();
	private ArrayList<String> featureTruth = null;

	HashMap<String, Integer> loadAnnotations(String annotFile) throws IOException {
		HashMap<String, Integer> list = new HashMap<String, Integer>();
		List<String> lines = Files.readAllLines(Paths.get(annotFile));
		
		for (String line : lines) {
			if (line.startsWith("#"))
				continue;
			String[] split = line.split("\t");
			String mention = split[3];
			String streamid = split[2];
			Integer annotation = Integer.decode(split[5]);

			list.put(streamid + "\t" + mention, annotation);
		}
		
		return list;
	}

	
	public HashMap<String, String> getFeatures(String vector) throws Exception {
		HashMap<String, String> lookup = new HashMap<String, String>();
		String[] split = vector.split("\t");
		for (String feat : split) {
			String[] nameVal;
			nameVal = feat.split(":");
			if (nameVal[1].equals("?"))
				nameVal[1] = "0"; // RankLib does not like unknown features
			lookup.put(nameVal[0], nameVal[1]);
		}

		return lookup;
	}
	
	public void processFeatureVector(String key, String featureVector) throws Exception {
		String[] split = key.split("\t");
		String entity = split[1];
		HashMap<String, String> featureLookup = getFeatures(featureVector);


		if (featureTruth == null) {
			featureTruth = new ArrayList<String>();
			// Feature vector of the first instance becomes defacto feature vector for the whole run
			featureTruth.addAll(featureLookup.keySet());
		}

		ArrayList<String> entityTestingVector = entityTestingVectors.get(entity);
		// New entity, we need to initialize it's vectors
		if (entityTestingVector == null) {
			String logString = "Detected new entity " + entity;
			System.out.println(logString);

			entityTrainingVectors.put(entity, new ArrayList<String>());
			entityTestingVectors.put(entity, new ArrayList<String>());
			entityTestingIds.put(entity, new ArrayList<String>());
			entityTestingVector = entityTestingVectors.get(entity);

		}

		// Get entity-document ids
		ArrayList<String> myTestingIds = entityTestingIds.get(entity);


		// Build the SVMRank feature vector for this instance
		String svmRankInst = " qid:1";
		for (int i  = 0; i < featureTruth.size(); i++) {
			String val = featureLookup.get(featureTruth.get(i));

			if (val == null) {
				System.out.println("Warning: Missing feature " + featureTruth.get(i) + " in " + key);
				val = "0";
			}

			svmRankInst += " " + (i+1) + ":" + val;
		}

		if (featureTruth.size() != featureLookup.size()) {
			System.out.println("Warning: Entity feature count not equal to instance for " + key);
		}

		entityTestingVector.add("0" + svmRankInst + " #" + key);
		myTestingIds.add(key);

		Integer training_rank = annotationsTraining.get(key);
		if (training_rank != null) {
			ArrayList<String> entityTrainingVector = entityTrainingVectors.get(entity);
			entityTrainingVector.add(Integer.toString(training_rank+1) + svmRankInst);
		}
	}

	private int runRankLibAndPrintErrors(String args) throws InterruptedException, IOException {
		Process pr = Runtime.getRuntime().exec("java -jar lib/RankLib-2.9.jar " + args); // RF ranker
		pr.waitFor();

		BufferedReader in = new BufferedReader(new InputStreamReader(pr.getErrorStream()));
		String line;
		while ((line = in.readLine()) != null) {
			System.out.println(line);
		}

		return pr.exitValue();
	}
	
	public void run(String annotFileTraining, String sequenceFiles, String outputDir, String runName, String rankerSettings) throws Exception {
		annotationsTraining = loadAnnotations(annotFileTraining);

		Configuration conf = new Configuration();
		FileSystem fs = FileSystem.get(conf);

		// Build lists of SVMRank instance lines for each entity
		for (FileStatus file : fs.listStatus(new Path(sequenceFiles))) {
			if (!file.getPath().getName().startsWith("part")) {
				continue;
			}
			Reader reader = new Reader(conf, Reader.file(file.getPath()));

			Text key = new Text();
			Text featureVector = new Text();

			while (reader.next(key, featureVector)) {
				processFeatureVector(key.toString(), featureVector.toString());
			}
			reader.close();
		}

		// Output SVMRank instance lines for each entity separately
		TrecKBARunWriter runWriter = new TrecKBARunWriter(outputDir, "ntnu", runName);

		String trainFileGlobal = outputDir + "/global_training.dat";
		BufferedWriter trainingWriterGlobal = new BufferedWriter(new FileWriter(trainFileGlobal));

		// Train the entity-dependent models
		HashMap<String, Float> scores = new HashMap<String, Float>();
		for (String entity : entityTrainingVectors.keySet()) {
			System.out.println("Processing " + entity + "...");
			String entityShortName = EntityNameResolver.removeNamePrefix(entity);
			String trainFileEntity = outputDir + "/" + entityShortName + "_training.dat";
			BufferedWriter trainingWriter = new BufferedWriter(new FileWriter(trainFileEntity));



			ArrayList<String> entityTrainingVector = entityTrainingVectors.get(entity);

			for (String inst : entityTrainingVector) {
				trainingWriter.write(inst + "\n");
				trainingWriterGlobal.write(inst + "\n");
			}
			trainingWriter.close();

			// Train entity-dependant model (If it has enough training instances)
			if (entityTrainingVector.size() > 50) {
				System.out.println("Training model on " + entityTrainingVector.size());
				String modelFile = outputDir + "/" + entityShortName + ".model";
				int retVal = runRankLibAndPrintErrors("-train " + trainFileEntity +
						" -save " + modelFile +
						" -gmax 1 " +
						rankerSettings);

				if (retVal != 0) {
					System.out.println("Failed to train for " + entity);
					return;
				}
			}

		}
		trainingWriterGlobal.close();

		// Train global model
		System.out.println("Training global model");
		int retVal = runRankLibAndPrintErrors("-train " + trainFileGlobal +
				" -save " + outputDir + "/global.model" +
				" -gmax 1 " +
				rankerSettings);

		if (retVal != 0) {
			System.out.println("Failed to train global model");
			return;
		}

		// Test for all entities
		for (String entity : entityTestingVectors.keySet()) {
			String entityShortName = EntityNameResolver.removeNamePrefix(entity);

			String testFile = outputDir + "/" + entityShortName + "_testing.dat";
			BufferedWriter testingWriter = new BufferedWriter(new FileWriter(testFile));
			ArrayList<String> myTestingIds = entityTestingIds.get(entity);
			ArrayList<String> entityTestingVector = entityTestingVectors.get(entity);
			ArrayList<String> entityTrainingVector = entityTrainingVectors.get(entity);

			for (String inst : entityTestingVector) {
				testingWriter.write(inst + "\n");
			}
			testingWriter.close();

			String modelFile = outputDir + "/" + entityShortName + ".model";
			if (entityTrainingVector.size() < 50) {
				modelFile = outputDir + "/global.model";

			}

			// Test for testing instances
			System.out.println("Testing model on " + entityTestingVector.size());
			String scoreFile = outputDir + "/" + entityShortName + ".scores";
			retVal = runRankLibAndPrintErrors("-load " + modelFile +
					" -rank " + testFile +
					" -score " + scoreFile +
					" -gmax 1");

			if (retVal != 0) {
				System.out.println("Failed to test for " + entity);
				return;
			}

			List<String> lines = Files.readAllLines(Paths.get(scoreFile));
			for(int i = 0; i < lines.size(); i++) {
				String line = lines.get(i);
				float score = Float.parseFloat(line.split("\t")[2]);
				String id = myTestingIds.get(i);

				scores.put(id, score);
			}
		}

		// Build scores from score file and calculate min/max values
		float min = Float.MAX_VALUE;
		float max = Float.MIN_VALUE;
		for (Float score : scores.values()) {
			if (score < min) {
				min = score;
			}
			if (score > max) {
				max = score;
			}
		}

		// Rescore according to min/max score
		for (Map.Entry<String, Float> entry : scores.entrySet()) {
			int score = Math.round((entry.getValue()-min)/(max-min)*1000);
			if (score == 0) {
				score = 1;
			}

			runWriter.writeJudgement(entry.getKey(), score, 2);
		}

		runWriter.close();

	}
	
	public static void main(String[] args) throws Exception {
		String sequenceFiles = null;
		String trainingAnnotFile = null;
		String outputDir = null;
		String runName = "default";
		String ranker = "-ranker 8";

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-a".equals(args[i])) {
					trainingAnnotFile = args[++i];
				} else if ("-i".equals(args[i])) {
					sequenceFiles = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				} else if ("-n".equals(args[i])) {
					runName = args[++i];
				} else if ("-s".equals(args[i])) {
					ranker = args[++i];
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
			}
		}

		if (trainingAnnotFile == null || outputDir == null || sequenceFiles == null) {
			System.out.println("Missing input, aborting");
			return;
		}

		
		RankLibRanker a = new RankLibRanker();
		
		a.run(trainingAnnotFile, sequenceFiles, outputDir, runName, ranker);
	}
}
