package no.ntnu.utils.classification;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.SequenceFile.Reader;
import org.apache.hadoop.io.Text;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.*;

/**
 * Use Ranklib to train entity-independent ranking model and produce runfile
 *
 * Ranklib must be located in lib/RankLib-2.9.jar
 *
 * Parameters:
 * a: Path to file containing training data
 * i: Path to file containing entity-document feature vectors
 * n: Run name
 * s: Ranklib settings (Default Random Forests with default hyperparams)
 *
 * @author Christian Barth Roligheten <chrbarrol@gmail.com>
 *
 */

public class RankLibRankerSingle {
	private HashMap<String, Integer> annotationsTraining;

	private HashMap<String, ArrayList<String>> entityTrainingVectors = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<String>> entityTestingVectors = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<String>> entityTestingIds = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<String>> entityVectors = new HashMap<String, ArrayList<String>>();
	private HashMap<String, Integer> entityIds = new HashMap<String, Integer>();
	private int idCounter = 1;

	HashMap<String, Integer> loadAnnotations(String annotFile) throws IOException {
		HashMap<String, Integer> list = new HashMap<String, Integer>();
		List<String> lines = Files.readAllLines(Paths.get(annotFile));
		
		for (String line : lines) {
			if (line.startsWith("#"))
				continue;
			String[] split = line.split("\t");
			String mention = split[3];
			String streamid = split[2];
			Integer annotation = Integer.decode(split[5]);

			list.put(streamid + "\t" + mention, annotation);
		}
		
		return list;
	}

	
	public HashMap<String, String> getFeatures(String vector) throws Exception {
		HashMap<String, String> lookup = new HashMap<String, String>();
		String[] split = vector.split("\t");
		for (String feat : split) {
			String[] nameVal;
			nameVal = feat.split(":");
			if (nameVal[1].equals("?") || nameVal[1].equals("NaN"))
				nameVal[1] = "0"; // RankLib does not like unknown features
			lookup.put(nameVal[0], nameVal[1]);
		}

		return lookup;
	}
	
	public void processFeatureVector(String key, String featureVector, List<String> exclude) throws Exception {
		String[] split = key.split("\t");
		String entity = split[1];
		HashMap<String, String> featureLookup = getFeatures(featureVector);

		Integer myId = entityIds.get(entity);
		if (myId == null) {
			myId = idCounter;
			entityIds.put(entity, myId);
			idCounter++;
		}

		ArrayList<String> entityVector = entityVectors.get(entity);

		// New entity, we need to initialize it's vectors
		if (entityVector == null) {
			entityVector = new ArrayList<String>();
			// Feature vector of the first instance becomes defacto feature vector for the whole run
			for (String feat : featureLookup.keySet()) {
				if (!exclude.contains(feat)) { // Remove excluded features
					entityVector.add(feat);
				}
			}

			String logString = "Detected new entity " + entity + ". Using " + entityVector.size() + " features.";
			System.out.println(logString);

			entityVectors.put(entity, entityVector);

			entityTrainingVectors.put(entity, new ArrayList<String>());
			entityTestingVectors.put(entity, new ArrayList<String>());
			entityTestingIds.put(entity, new ArrayList<String>());
		}

		// Get entity's training and testing vectors
		ArrayList<String> entityTestingVector = entityTestingVectors.get(entity);
		ArrayList<String> entityTestingId = entityTestingIds.get(entity);


		// Build the SVMRank feature vector for this instance
		String svmRankInst = " qid:" + myId;
		for (int i  = 0; i < entityVector.size(); i++) {
			String val = featureLookup.get(entityVector.get(i));

			if (val == null) {
				System.out.println("Warning: Missing feature " + entityVector.get(i) + " in " + key);
				val = "0";
			}

			svmRankInst += " " + (i+1) + ":" + val;
		}

		if (entityVector.size() != (featureLookup.size()-exclude.size())) {
			System.out.println("Warning: Entity feature count not equal to instance for " + key);
		}

		entityTestingVector.add("0" + svmRankInst);
		entityTestingId.add(key);

		Integer training_rank = annotationsTraining.get(key);

		if (training_rank != null) {
			ArrayList<String> entityTrainingVector = entityTrainingVectors.get(entity);
			entityTrainingVector.add(Integer.toString(training_rank+1) + svmRankInst);

			boolean isVital = training_rank == 2;
		}
	}

	private int runRankLibAndPrintErrors(String args) throws InterruptedException, IOException {
		Process pr = Runtime.getRuntime().exec("java -jar lib/RankLib-2.9.jar " + args); // RF ranker
		pr.waitFor();

		BufferedReader in = new BufferedReader(new InputStreamReader(pr.getErrorStream()));
		String line;
		while ((line = in.readLine()) != null) {
			System.out.println(line);
		}

		return pr.exitValue();
	}
	
	public void run(String annotFileTraining, String sequenceFiles, String outputDir, String runName, List<String> exclude, int ranker) throws Exception {
		annotationsTraining = loadAnnotations(annotFileTraining);

		Configuration conf = new Configuration();
		FileSystem fs = FileSystem.get(conf);

		// Build lists of SVMRank instance lines for each entity
		for (FileStatus file : fs.listStatus(new Path(sequenceFiles))) {
			if (!file.getPath().getName().startsWith("part")) {
				continue;
			}
			Reader reader = new Reader(conf, Reader.file(file.getPath()));

			Text key = new Text();
			Text featureVector = new Text();

			while (reader.next(key, featureVector)) {
				processFeatureVector(key.toString(), featureVector.toString(), exclude);
			}
			reader.close();
		}

		// Output SVMRank instance lines for each entity separately
		TrecKBARunWriter runWriter = new TrecKBARunWriter(outputDir, "ntnu", runName);

		String trainFile = outputDir + "/" + "training.dat";
		String testFile = outputDir + "/" + "testing.dat";
		BufferedWriter trainingWriter = new BufferedWriter(new FileWriter(trainFile));
		BufferedWriter testingWriter = new BufferedWriter(new FileWriter(testFile));

		HashMap<String, Float> scores = new HashMap<String, Float>();

		ArrayList<String> testingIds = new ArrayList<String>();
		for (String entity : entityVectors.keySet()) {
			System.out.println("Processing " + entity + "...");


			ArrayList<String> entityTrainingVector = entityTrainingVectors.get(entity);
			ArrayList<String> entityTestingVector = entityTestingVectors.get(entity);
			ArrayList<String> entityIds = entityTestingIds.get(entity);


			for (String inst : entityTestingVector) {
				testingWriter.write(inst + "\n");
			}
			for (String key : entityIds) {
				testingIds.add(key);
			}
			for (String inst : entityTrainingVector) {
				trainingWriter.write(inst + "\n");
			}
		}

		trainingWriter.close();
		testingWriter.close();

		// Train entity-dependant model
		String modelFile = outputDir + "/" + "model.model";
		int retVal = runRankLibAndPrintErrors("-train " + trainFile +
				" -save " + modelFile +
				" -ranker " + Integer.toString(ranker) +
				" -gmax 3");

		if (retVal != 0) {
			System.out.println("Failed to train");
			return;
		}

		// Test for testing instances
		String scoreFile = outputDir + "/" + "model.scores";
		retVal = runRankLibAndPrintErrors("-load " + modelFile +
				" -rank " + testFile +
				" -score " + scoreFile +
				" -gmax 3");

		if (retVal != 0) {
			System.out.println("Failed to test");
			return;
		}

		List<String> lines = Files.readAllLines(Paths.get(scoreFile));
		for(int i = 0; i < lines.size(); i++) {
			String line = lines.get(i);
			float score = Float.parseFloat(line.split("\t")[2]);
			String id = testingIds.get(i);

			scores.put(id, score);
		}

		// Build scores from score file and calculate min/max values
		float min = Float.MAX_VALUE;
		float max = Float.MIN_VALUE;
		for (Float score : scores.values()) {
			if (score < min) {
				min = score;
			}
			if (score > max) {
				max = score;
			}
		}

		// Rescore according to min/max score
		for (Map.Entry<String, Float> entry : scores.entrySet()) {
			int score = Math.round((entry.getValue()-min)/(max-min)*1000);
			if (score == 0) {
				score = 1;
			}

			runWriter.writeJudgement(entry.getKey(), score, 2);
		}

		runWriter.close();

	}
	
	public static void main(String[] args) throws Exception {
		String sequenceFiles = null;
		String trainingAnnotFile = null;
		String outputDir = null;
		String runName = "default";
		int ranker = 8;
		List<String> exclude = new ArrayList<String>();

		// parse arguments
		for (int i = 0; i < args.length; ++i) {
			try {
				if ("-a".equals(args[i])) {
					trainingAnnotFile = args[++i];
				} else if ("-i".equals(args[i])) {
					sequenceFiles = args[++i];
				} else if ("-o".equals(args[i])) {
					outputDir = args[++i];
				} else if ("-n".equals(args[i])) {
					runName = args[++i];
				} else if ("-e".equals(args[i])) {
					exclude.addAll(Arrays.asList(args[++i].split(",")));
				} else if ("-r".equals(args[i])) {
					ranker = Integer.parseInt(args[++i]);
				}
			} catch (ArrayIndexOutOfBoundsException except) {
				System.out.println("ERROR: Required parameter missing from "
						+ args[i - 1]);
			}
		}

		if (trainingAnnotFile == null || outputDir == null || sequenceFiles == null) {
			System.out.println("Missing input, aborting");
			return;
		}

		
		RankLibRankerSingle a = new RankLibRankerSingle();
		
		a.run(trainingAnnotFile, sequenceFiles, outputDir, runName, exclude, ranker);
	}
}
