package no.ntnu.wikipedia;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

/**
 * Process Wikipedia history files
 * 
 * Starting point is a set of Wikipedia XML files where we grep out <title>, <id>, and <timestamp> lines
 * 
 *    <title>User:Jeronimo/Sandbox</title>
 *    <id>1014632</id>
 *      <id>6126857</id>
 *      <timestamp>2004-09-24T19:13:12Z</timestamp>
 *        <id>108</id>
 *      <id>6130144</id>
 *      <timestamp>2004-09-24T19:18:14Z</timestamp>
 *        <id>108</id>
 *      <id>6139882</id>
 *      <timestamp>2004-09-24T22:05:46Z</timestamp>
 *        <id>108</id>
 *
 * This is then converted to a slightly more space-efficient format (.out files): 
 * 
 *    title:User:Jeronimo/Sandbox
 *    id:1014632
 *      id:6126857
 *      t:2004-09-24T19:13:12Z
 *        id:108
 *      id:6130144
 *      t:2004-09-24T19:18:14Z
 *        id:108
 *      id:6139882
 *      t:2004-09-24T22:05:46Z
 * 
 * The output is one line per entity with all timestamps:
 * Title \t PageId \t Timestamps-space-separated
 * 
 * I.e., 
 * User:Jeronimo/Sandbox	1014632	2004-09-24T19:13:12Z 2004-09-24T19:18:14Z 2004-09-24T22:05:46Z ...
 *
 * @author Krisztian Balog <krisztian.balog@uis.no>
 */
public class ProcessWPHistory {

	/**
	 * Process a single .out file
	 * Note that the number of spaces matter!
	 * 
	 * @param inputFile
	 * @param output
	 */
	private static void processFile(String inputFile, Writer output) {
		System.out.print("Processing " + inputFile + " ... ");
		int num = 0;
		try {
			// write feature vectors
			BufferedReader br = new BufferedReader(new FileReader(inputFile));
			String line;
			String title = null;
			String id = null;
			String timeStamp = "";
			while ((line = br.readLine()) != null) {

				if (line.startsWith("    title:")) {
					// write out previous record
					if (title != null) {
						output.write(title + "\t" + id + "\t" + timeStamp + "\n");
						num++;
						title = null;
						id = null;
						timeStamp = "";
					}
					
					title = line.substring(10);
					id = null;
				}
				else if (line.startsWith("    id:")) {
					if (id == null && title != null) {
						id = line.substring(7);
					}
					else {
						System.err.println("Input error! Skipping entity!");
					}
				}
				else if (line.startsWith("      t:")) {
					timeStamp += line.substring(8) + " ";
				}								
			}
			// write out last record
			if (title != null) {
				output.write(title + "\t" + id + "\t" + timeStamp + "\n");
				num++;
			}
			
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}		
		System.out.println(num + " entities processed");
	}
	
	private static void run(String inputDir, String outputFile) {

		try {
			Writer output = new BufferedWriter(new FileWriter(outputFile));
			// list all files in inputDir
			for (File file : new File(inputDir).listFiles()) {
				String fileName = file.getName();
				if (fileName.endsWith(".out")) {
					processFile(file.getAbsolutePath(), output);
				}
			}
			output.close();
		} catch (IOException e) {
			e.printStackTrace();
		}		
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		if (args.length != 2) {
			System.out.println("Usage: " + ProcessWPHistory.class + " inputDir outputFile");
			System.exit(-1);
		}
		
		String inputDir = args[0];
		String outputFile = args[1];

		run(inputDir, outputFile);
	}

}
